"""
The ImageData definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


from typing import List, Union, Optional

import numpy

from sarpy.io.xml.base import Serializable, Arrayable, SerializableArray
from sarpy.io.xml.descriptors import IntegerDescriptor, FloatArrayDescriptor, \
    StringEnumDescriptor, SerializableDescriptor, SerializableArrayDescriptor
from sarpy.geometry.geometry_elements import LinearRing

from .base import DEFAULT_STRICT, FLOAT_FORMAT
from .blocks import RowColType, RowColArrayElement


class FullImageType(Serializable, Arrayable):
    """
    The full image product attributes.
    """

    _fields = ('NumRows', 'NumCols')
    _required = _fields
    # descriptors
    NumRows = IntegerDescriptor(
        'NumRows', _required, strict=DEFAULT_STRICT,
        docstring='Number of rows in the original full image product. May include zero pixels.')  # type: int
    NumCols = IntegerDescriptor(
        'NumCols', _required, strict=DEFAULT_STRICT,
        docstring='Number of columns in the original full image product. May include zero pixels.')  # type: int

    def __init__(
            self,
            NumRows: int = None,
            NumCols: int = None,
            **kwargs):
        """

        Parameters
        ----------
        NumRows : int
        NumCols : int
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.NumRows, self.NumCols = NumRows, NumCols
        super(FullImageType, self).__init__(**kwargs)

    def get_array(self, dtype=numpy.int64) -> numpy.ndarray:
        """Gets an array representation of the class instance.

        Parameters
        ----------
        dtype : str|numpy.dtype|numpy.number
            numpy data type of the return

        Returns
        -------
        numpy.ndarray
            array of the form `[X,Y,Z]`
        """

        return numpy.array([self.NumRows, self.NumCols], dtype=dtype)

    @classmethod
    def from_array(cls, array: Union[numpy.ndarray, list, tuple]):
        """
        Create from an array type entry.

        Parameters
        ----------
        array: numpy.ndarray|list|tuple
            assumed `[NumRows, NumCols]`

        Returns
        -------
        FullImageType
        """

        if isinstance(array, (numpy.ndarray, list, tuple)):
            if len(array) < 2:
                raise ValueError('Expected array to be of length 2, and received {}'.format(len(array)))
            return cls(NumRows=array[0], NumCols=array[1])
        raise ValueError('Expected array to be numpy.ndarray, list, or tuple, got {}'.format(type(array)))


class ImageDataType(Serializable):
    """The image pixel data."""
    _collections_tags = {
        'AmpTable': {'array': True, 'child_tag': 'Amplitude'},
        'ValidData': {'array': True, 'child_tag': 'Vertex'},
    }
    _fields = (
        'PixelType', 'AmpTable', 'NumRows', 'NumCols', 'FirstRow', 'FirstCol', 'FullImage', 'SCPPixel', 'ValidData')
    _required = ('PixelType', 'NumRows', 'NumCols', 'FirstRow', 'FirstCol', 'FullImage', 'SCPPixel')
    _numeric_format = {'AmpTable': FLOAT_FORMAT}
    _PIXEL_TYPE_VALUES = ("RE32F_IM32F", "RE16I_IM16I", "AMP8I_PHS8I")
    # descriptors
    PixelType = StringEnumDescriptor(
        'PixelType', _PIXEL_TYPE_VALUES, _required, strict=True,
        docstring="The PixelType attribute which specifies the interpretation of the file data.")  # type: str
    AmpTable = FloatArrayDescriptor(
        'AmpTable', _collections_tags, _required, strict=DEFAULT_STRICT,
        minimum_length=256, maximum_length=256,
        docstring="The amplitude look-up table. This is required if "
                  "`PixelType == 'AMP8I_PHS8I'`")  # type: numpy.ndarray
    NumRows = IntegerDescriptor(
        'NumRows', _required, strict=True,
        docstring='The number of Rows in the product. May include zero rows.')  # type: int
    NumCols = IntegerDescriptor(
        'NumCols', _required, strict=True,
        docstring='The number of Columns in the product. May include zero rows.')  # type: int
    FirstRow = IntegerDescriptor(
        'FirstRow', _required, strict=DEFAULT_STRICT,
        docstring='Global row index of the first row in the product. '
                  'Equal to 0 in full image product.')  # type: int
    FirstCol = IntegerDescriptor(
        'FirstCol', _required, strict=DEFAULT_STRICT,
        docstring='Global column index of the first column in the product. '
                  'Equal to 0 in full image product.')  # type: int
    FullImage = SerializableDescriptor(
        'FullImage', FullImageType, _required, strict=DEFAULT_STRICT,
        docstring='Original full image product.')  # type: FullImageType
    SCPPixel = SerializableDescriptor(
        'SCPPixel', RowColType, _required, strict=DEFAULT_STRICT,
        docstring='Scene Center Point pixel global row and column index. Should be located near the '
                  'center of the full image.')  # type: RowColType
    ValidData = SerializableArrayDescriptor(
        'ValidData', RowColArrayElement, _collections_tags, _required, strict=DEFAULT_STRICT, minimum_length=3,
        docstring='Indicates the full image includes both valid data and some zero filled pixels. '
                  'Simple polygon encloses the valid data (may include some zero filled pixels for simplification). '
                  'Vertices in clockwise order.')  # type: Union[SerializableArray, List[RowColArrayElement]]

    def __init__(
            self,
            PixelType: str = None,
            AmpTable: Optional[numpy.ndarray] = None,
            NumRows: int = None,
            NumCols: int = None,
            FirstRow: int = None,
            FirstCol: int = None,
            FullImage: Union[FullImageType, numpy.ndarray, list, tuple] = None,
            SCPPixel: Union[RowColType, numpy.ndarray, list, tuple] = None,
            ValidData=None,
            **kwargs):
        """

        Parameters
        ----------
        PixelType : str
        AmpTable : numpy.ndarray|list|tuple
        NumRows : int
        NumCols : int
        FirstRow : int
        FirstCol : int
        FullImage : FullImageType|numpy.ndarray|list|tuple
        SCPPixel : RowColType|numpy.ndarray|list|tuple
        ValidData : SerializableArray|List[RowColArrayElement]|numpy.ndarray|list|tuple
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.PixelType = PixelType
        self.AmpTable = AmpTable
        self.NumRows, self.NumCols = NumRows, NumCols
        self.FirstRow, self.FirstCol = FirstRow, FirstCol
        self.FullImage = FullImage
        self.SCPPixel = SCPPixel
        self.ValidData = ValidData
        super(ImageDataType, self).__init__(**kwargs)

    def _check_valid_data(self) -> bool:
        if self.ValidData is None:
            return True
        if len(self.ValidData) < 2:
            return True

        value = True
        valid_data = self.ValidData.get_array(dtype='float64')
        lin_ring = LinearRing(coordinates=valid_data)
        area = lin_ring.get_area()
        if area == 0:
            self.log_validity_error('ValidData encloses no area.')
            value = False
        elif area > 0:
            self.log_validity_error(
                "ValidData must be traversed in clockwise direction.")
            value = False
        for i, entry in enumerate(valid_data):
            if not ((self.FirstRow <= entry[0] <= self.FirstRow + self.NumRows) and
                    (self.FirstCol <= entry[1] <= self.FirstCol + self.NumCols)):
                self.log_validity_warning(
                    'ValidData entry {} is not contained in the image bounds'.format(i))
                value = False
        return value

    def _basic_validity_check(self) -> bool:
        condition = super(ImageDataType, self)._basic_validity_check()
        if (self.PixelType == 'AMP8I_PHS8I') and (self.AmpTable is None):
            self.log_validity_error(
                "We have `PixelType='AMP8I_PHS8I'` and `AmpTable` is not defined for ImageDataType.")
            condition = False
        if (self.PixelType != 'AMP8I_PHS8I') and (self.AmpTable is not None):
            self.log_validity_error(
                "We have `PixelType != 'AMP8I_PHS8I'` and `AmpTable` is defined for ImageDataType.")
            condition = False
        if (self.ValidData is not None) and (len(self.ValidData) < 3):
            self.log_validity_error(
                "We have `ValidData` defined with fewer than 3 entries.")
            condition = False
        condition &= self._check_valid_data()
        return condition

    def get_valid_vertex_data(self, dtype=numpy.int64) -> Optional[numpy.ndarray]:
        """
        Gets an array of `[row, col]` indices defining the valid data. If this is not viable, then `None`
        will be returned.

        Parameters
        ----------
        dtype : object
            the data type for the array

        Returns
        -------
        numpy.ndarray|None
        """

        if self.ValidData is None:
            return None
        out = numpy.zeros((self.ValidData.size, 2), dtype=dtype)
        for i, entry in enumerate(self.ValidData):
            out[i, :] = entry.get_array(dtype=dtype)
        return out

    def get_full_vertex_data(self, dtype=numpy.int64) -> Optional[numpy.ndarray]:
        """
        Gets an array of `[row, col]` indices defining the full vertex data. If this is not viable, then `None`
        will be returned.

        Parameters
        ----------
        dtype : object
            the data type for the array

        Returns
        -------
        numpy.ndarray|None
        """

        if self.NumRows is None or self.NumCols is None:
            return None
        return numpy.array(
            [[0, 0], [0, self.NumCols - 1], [self.NumRows - 1, self.NumCols - 1], [self.NumRows - 1, 0]], dtype=dtype)

    def get_pixel_size(self) -> int:
        """
        Gets the size per pixel, in bytes.

        Returns
        -------
        int
        """

        if self.PixelType == "RE32F_IM32F":
            return 8
        elif self.PixelType == "RE16I_IM16I":
            return 4
        elif self.PixelType == "AMP8I_PHS8I":
            return 2
        else:
            raise ValueError('Got unhandled pixel type `{}`'.format(self.PixelType))



"""
The TimelineType definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


from typing import List, Union, Optional
from datetime import datetime, date

import numpy

from sarpy.io.xml.base import Serializable, SerializableArray
from sarpy.io.xml.descriptors import FloatDescriptor, IntegerDescriptor, \
    DateTimeDescriptor, SerializableDescriptor, SerializableArrayDescriptor

from .base import DEFAULT_STRICT, FLOAT_FORMAT
from .blocks import Poly1DType


class IPPSetType(Serializable):
    """
    The Inter-Pulse Parameter array element container.
    """

    # NOTE that this is simply defined as a child class ("Set") of the TimelineType in the SICD standard
    #   Defining it at root level clarifies the documentation, and giving it a more descriptive name is
    #   appropriate.
    _fields = ('TStart', 'TEnd', 'IPPStart', 'IPPEnd', 'IPPPoly', 'index')
    _required = _fields
    _set_as_attribute = ('index', )
    _numeric_format = {'TStart': FLOAT_FORMAT, 'TEnd': FLOAT_FORMAT, }
    # descriptors
    TStart = FloatDescriptor(
        'TStart', _required, strict=DEFAULT_STRICT,
        docstring='IPP start time relative to collection start time, i.e. offsets in seconds.')  # type: float
    TEnd = FloatDescriptor(
        'TEnd', _required, strict=DEFAULT_STRICT,
        docstring='IPP end time relative to collection start time, i.e. offsets in seconds.')  # type: float
    IPPStart = IntegerDescriptor(
        'IPPStart', _required, strict=True, docstring='Starting IPP index for the period described.')  # type: int
    IPPEnd = IntegerDescriptor(
        'IPPEnd', _required, strict=True, docstring='Ending IPP index for the period described.')  # type: int
    IPPPoly = SerializableDescriptor(
        'IPPPoly', Poly1DType, _required, strict=DEFAULT_STRICT,
        docstring='IPP index polynomial coefficients yield IPP index as a function of time.')  # type: Poly1DType
    index = IntegerDescriptor(
        'index', _required, strict=DEFAULT_STRICT, docstring='The element array index.')  # type: int

    def __init__(
            self,
            TStart: float = None,
            TEnd: float = None,
            IPPStart: int = None,
            IPPEnd: int = None,
            IPPPoly: Union[Poly1DType, numpy.ndarray, list, tuple] = None,
            index: int = None,
            **kwargs):
        """

        Parameters
        ----------
        TStart : float
        TEnd : float
        IPPStart : int
        IPPEnd : int
        IPPPoly : Poly1DType|numpy.ndarray|list|tuple
        index : int
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.TStart, self.TEnd = TStart, TEnd
        self.IPPStart, self.IPPEnd = IPPStart, IPPEnd
        self.IPPPoly = IPPPoly
        self.index = index
        super(IPPSetType, self).__init__(**kwargs)

    def _basic_validity_check(self) -> bool:
        condition = super(IPPSetType, self)._basic_validity_check()
        if self.TStart >= self.TEnd:
            self.log_validity_error(
                'TStart ({}) >= TEnd ({})'.format(self.TStart, self.TEnd))
            condition = False

        if self.IPPStart >= self.IPPEnd:
            self.log_validity_error(
                'IPPStart ({}) >= IPPEnd ({})'.format(self.IPPStart, self.IPPEnd))
            condition = False

        ipp_start_from_poly = round(self.IPPPoly(self.TStart))
        if self.IPPStart != ipp_start_from_poly:
            self.log_validity_error(
                f'IPPStart ({self.IPPStart}) inconsistent with IPPPoly(TStart) ({ipp_start_from_poly})'
            )
            condition = False

        ipp_end_from_poly = round(self.IPPPoly(self.TEnd) - 1)
        if self.IPPEnd != ipp_end_from_poly:
            self.log_validity_error(
                f'IPPEnd ({self.IPPEnd}) inconsistent with IPPPoly(TEnd) - 1 ({ipp_end_from_poly})'
            )
            condition = False

        prf = self.IPPPoly.derivative_eval((self.TStart + self.TEnd)/2)
        if prf < 0:
            self.log_validity_error(f'IPPSet has a negative PRF: {prf}')
            condition = False
        if prf > 100e3:
            self.log_validity_warning(f'IPPSet has an unreasonable PRF: {prf}')

        return condition


class TimelineType(Serializable):
    """
    The details for the imaging collection timeline.
    """

    _fields = ('CollectStart', 'CollectDuration', 'IPP')
    _required = ('CollectStart', 'CollectDuration', )
    _collections_tags = {'IPP': {'array': True, 'child_tag': 'Set'}}
    _numeric_format = {'CollectDuration': FLOAT_FORMAT, }
    # descriptors
    CollectStart = DateTimeDescriptor(
        'CollectStart', _required, strict=DEFAULT_STRICT, numpy_datetime_units='us',
        docstring='The collection start time. The default precision will be microseconds.')  # type: numpy.datetime64
    CollectDuration = FloatDescriptor(
        'CollectDuration', _required, strict=DEFAULT_STRICT,
        docstring='The duration of the collection in seconds.')  # type: float
    IPP = SerializableArrayDescriptor(
        'IPP', IPPSetType, _collections_tags, _required, strict=DEFAULT_STRICT, minimum_length=1,
        docstring="The Inter-Pulse Period (IPP) parameters array.")  # type: Union[SerializableArray, List[IPPSetType]]

    def __init__(
            self,
            CollectStart: Union[numpy.datetime64, datetime, date, str] = None,
            CollectDuration: float = None,
            IPP: Union[None, SerializableArray, List[IPPSetType]] = None,
            **kwargs):
        """

        Parameters
        ----------
        CollectStart : numpy.datetime64|datetime|date|str
        CollectDuration : float
        IPP : None|List[IPPSetType]
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.CollectStart = CollectStart
        self.CollectDuration = CollectDuration
        self.IPP = IPP
        super(TimelineType, self).__init__(**kwargs)

    @property
    def CollectEnd(self) -> Optional[numpy.datetime64]:
        """
        None|numpy.datetime64: The collection end time, inferred from `CollectEnd` and `CollectDuration`,
        provided that both are populated.
        """

        if self.CollectStart is None or self.CollectDuration is None:
            return None

        return self.CollectStart + numpy.timedelta64(int(self.CollectDuration*1e6), 'us')

    def _check_ipp_consecutive(self) -> bool:
        if self.IPP is None or len(self.IPP) < 2:
            return True
        cond = True

        sorted_ipps = sorted(self.IPP, key=lambda x: x.index)
        tstarts = [x.TStart for x in sorted_ipps]
        tends = [x.TEnd for x in sorted_ipps]
        ippstarts = [x.IPPStart for x in sorted_ipps]
        ippends = [x.IPPEnd for x in sorted_ipps]
        if tstarts != sorted(tstarts):
            self.log_validity_error(f'The IPPSets are not in start time order. TStart: {tstarts}')
            cond = False

        if tends != sorted(tends):
            self.log_validity_error(f'The IPPSets are not in end time order. TEnd: {tends}')
            cond = False

        actual_indices = [x.index for x in sorted_ipps]
        if not numpy.array_equal(numpy.arange(1, len(self.IPP)+1), actual_indices):
            self.log_validity_error(f'IPPSets indices ({actual_indices}) are not 1..size; '
                                    'unable to perform adjacent IPPSet checks')
            return False

        tgaps = [ts - te for ts, te in zip(tstarts[1:], tends[:-1])]
        for ig, g in enumerate(tgaps):
            if g > 0:
                self.log_validity_error(f'There is a gap between IPPSet[index={ig+1}] and '
                                        f'IPPSet[index={ig+2}] of {g} seconds')
                cond = False
            if g < 0:
                self.log_validity_error(f'There is overlap between IPPSet[index={ig+1}] and '
                                        f'IPPSet[index={ig+2}] of {-g} seconds')
                cond = False

        igaps = [i_s - i_e for i_s, i_e in zip(ippstarts[1:], ippends[:-1])]
        for ig, g in enumerate(igaps):
            if g > 1:
                self.log_validity_error(f'There is a gap between IPPSet[index={ig+1}] and '
                                        f'IPPSet[index={ig+2}] of {g-1} IPPs')
                cond = False
            if g < 1:
                self.log_validity_error(f'There is overlap between IPPSet[index={ig+1}] and '
                                        f'IPPSet[index={ig+2}] of {1-g} IPPs')
                cond = False

        return cond

    def _check_ipp_times(self) -> bool:
        if self.IPP is None:
            return True

        cond = True
        min_time = min(x.TStart for x in self.IPP)
        max_time = max(x.TEnd for x in self.IPP)
        if min_time < 0:
            self.log_validity_error(f'Earliest TStart is negative: {min_time}')
            cond = False
        if not numpy.isclose(max_time - min_time, self.CollectDuration, atol=1e-2):
            self.log_validity_error(
                'time range in IPP entries ({}) not in keeping with populated '
                'CollectDuration ({})'.format(max_time-min_time, self.CollectDuration))
            cond = False
        return cond

    def _basic_validity_check(self) -> bool:
        condition = super(TimelineType, self)._basic_validity_check()
        condition &= self._check_ipp_consecutive()
        condition &= self._check_ipp_times()
        return condition



"""
Types implementing the SIDD 3.0 DigitalElevationData Parameters
"""
__classification__ = 'UNCLASSIFIED'

# SIDD 3.0 reuses the SIDD 2.0 types.  Make those symbols available in this module.
from sarpy.io.product.sidd2_elements.DigitalElevationData import (
    GeographicCoordinatesType,
    GeopositioningType,
    AccuracyType,
    PositionalAccuracyType,
    DigitalElevationDataType,
)

__REUSED__ = (  # to avoid unused import lint errors
    GeographicCoordinatesType,
    GeopositioningType,
    AccuracyType,
    PositionalAccuracyType,
    DigitalElevationDataType,
)




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class PT(TREElement):
    def __init__(self, value):
        super(PT, self).__init__()
        self.add_field('LON', 's', 15, value)
        self.add_field('LAT', 's', 15, value)


class ACHZ(TREElement):
    def __init__(self, value):
        super(ACHZ, self).__init__()
        self.add_field('UNIAAH', 's', 3, value)
        if self.UNIAAH != '':
            self.add_field('AAH', 's', 5, value)
        self.add_field('UNIAPH', 's', 3, value)
        if self.UNIAPH != '':
            self.add_field('APH', 's', 5, value)
        self.add_field('NUMPTS', 'd', 3, value)
        self.add_loop('PTs', self.NUMPTS, PT, value)


class ACCHZBType(TREElement):
    def __init__(self, value):
        super(ACCHZBType, self).__init__()
        self.add_field('NUMACHZ', 'd', 2, value)
        self.add_loop('ACHZs', self.NUMACHZ, ACHZ, value)


class ACCHZB(TREExtension):
    _tag_value = 'ACCHZB'
    _data_type = ACCHZBType



"""
Functionality for an aggregate sicd type reader, for opening multiple sicd type
files as a single reader object.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


from typing import Tuple, Sequence, Union

from sarpy.io.complex.converter import open_complex
from sarpy.io.general.base import AggregateReader, SarpyIOError
from sarpy.io.complex.base import SICDTypeReader
from sarpy.io.complex.sicd_elements.SICD import SICDType


class AggregateComplexReader(AggregateReader, SICDTypeReader):
    """
    Aggregate multiple sicd type readers into a single reader instance.

    **Changed in version 1.3.0** for reading changes.
    """

    __slots__ = ('_readers', '_index_mapping')

    def __init__(self, readers: Union[Sequence[str], Sequence[SICDTypeReader]]):
        """

        Parameters
        ----------
        readers : Sequence[str]|Sequence[SICDTypeReader]
        """

        readers = self._validate_readers(readers)
        AggregateReader.__init__(self, readers)
        sicds = self._define_sicds()
        SICDTypeReader.__init__(self, None, sicds)
        self._check_sizes()

    @staticmethod
    def _validate_readers(readers: Sequence[SICDTypeReader]) -> Tuple[SICDTypeReader, ...]:
        """
        Validate the input reader/file collection.

        Parameters
        ----------
        readers : Sequence[SICDTypeReader]

        Returns
        -------
        Tuple[SICDTypeReader]
        """

        if not isinstance(readers, Sequence):
            raise TypeError('input argument must be a list or tuple of readers/files. Got type {}'.format(type(readers)))

        # get a reader for each entry, and make sure that they are sicd type

        # validate each entry
        the_readers = []
        for i, entry in enumerate(readers):
            if isinstance(entry, str):
                try:
                    reader = open_complex(entry)
                except SarpyIOError:
                    raise SarpyIOError(
                        'Attempted and failed to open {} (entry {} of the input argument) '
                        'using the complex opener.'.format(entry, i))
            else:
                reader = entry

            if not isinstance(reader, SICDTypeReader):
                raise ValueError(
                    'Entry {} of the input argument does not correspond to a SICDTypeReader instance. '
                    'Got type {}.'.format(i, type(reader)))
            the_readers.append(reader)
        return tuple(the_readers)

    def _define_sicds(self) -> Tuple[SICDType, ...]:
        sicds = []
        for reader_index, sicd_index in self.index_mapping:
            reader = self._readers[reader_index]
            sicd = reader.get_sicds_as_tuple()[sicd_index]
            sicds.append(sicd)
        return tuple(sicds)




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class EXOPTAType(TREElement):
    def __init__(self, value):
        super(EXOPTAType, self).__init__()
        self.add_field('ANGLETONORTH', 's', 3, value)
        self.add_field('MEANGSD', 's', 5, value)
        self.add_field('RESERV01', 's', 1, value)
        self.add_field('DYNAMICRANGE', 's', 5, value)
        self.add_field('RESERV02', 's', 7, value)
        self.add_field('OBLANG', 's', 5, value)
        self.add_field('ROLLANG', 's', 6, value)
        self.add_field('PRIMEID', 's', 12, value)
        self.add_field('PRIMEBE', 's', 15, value)
        self.add_field('RESERV03', 's', 5, value)
        self.add_field('NSEC', 's', 3, value)
        self.add_field('RESERV04', 's', 2, value)
        self.add_field('RESERV05', 's', 7, value)
        self.add_field('NSEG', 's', 3, value)
        self.add_field('MAXLPSEG', 's', 6, value)
        self.add_field('RESERV06', 's', 12, value)
        self.add_field('SUNEL', 's', 5, value)
        self.add_field('SUNAZ', 's', 5, value)


class EXOPTA(TREExtension):
    _tag_value = 'EXOPTA'
    _data_type = EXOPTAType



"""
The Global type definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from typing import Union

import numpy

from sarpy.io.xml.base import Serializable, Arrayable
from sarpy.io.xml.descriptors import FloatDescriptor, DateTimeDescriptor, \
    StringEnumDescriptor, IntegerEnumDescriptor, SerializableDescriptor
from .base import DEFAULT_STRICT, FLOAT_FORMAT


class TimelineType(Serializable):
    """
    Parameters that describe the collection times for the data contained in the product.
    """

    _fields = ('CollectionStart', 'RcvCollectionStart', 'TxTime1', 'TxTime2')
    _required = ('CollectionStart', 'TxTime1', 'TxTime2')
    _numeric_format = {'TxTime1': FLOAT_FORMAT, 'TxTime2': FLOAT_FORMAT}
    # descriptors
    CollectionStart = DateTimeDescriptor(
        'CollectionStart', _required, strict=DEFAULT_STRICT, numpy_datetime_units='us',
        docstring='Collection Start date and time (UTC). Time reference used for times '
                  'measured from collection start (i.e. slow time t = 0). For bistatic '
                  'collections, the time is the transmit platform collection '
                  'start time. The default display precision is microseconds, but this '
                  'does not that accuracy in value.')  # type: numpy.datetime64
    RcvCollectionStart = DateTimeDescriptor(
        'RcvCollectionStart', _required, strict=DEFAULT_STRICT, numpy_datetime_units='us',
        docstring='Receive only platform collection date and start time.')  # type: numpy.datetime64
    TxTime1 = FloatDescriptor(
        'TxTime1', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Earliest TxTime value for any signal vector in the product. '
                  'Time relative to Collection Start in seconds.')  # type: float
    TxTime2 = FloatDescriptor(
        'TxTime2', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Latest TxTime value for any signal vector in the product. '
                  'Time relative to Collection Start in seconds.')  # type: float

    def __init__(self, CollectionStart=None, RcvCollectionStart=None, TxTime1=None, TxTime2=None, **kwargs):
        """

        Parameters
        ----------
        CollectionStart : numpy.datetime64|datetime|date|str
        RcvCollectionStart : None|numpy.datetime64|datetime|date|str
        TxTime1 : float
        TxTime2 : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.CollectionStart = CollectionStart
        self.RcvCollectionStart = RcvCollectionStart
        self.TxTime1 = TxTime1
        self.TxTime2 = TxTime2
        super(TimelineType, self).__init__(**kwargs)


class FxBandType(Serializable, Arrayable):
    """
    Parameters that describe the FX frequency limits for the signal array(s)
    contained in the product.
    """

    _fields = ('FxMin', 'FxMax')
    _required = _fields
    _numeric_format = {fld: FLOAT_FORMAT for fld in _fields}
    # descriptors
    FxMin = FloatDescriptor(
        'FxMin', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Minimum fx value for any signal vector in the product in '
                  'Hz.')  # type: float
    FxMax = FloatDescriptor(
        'FxMax', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Maximum fx value for any signal vector in the product in '
                  'Hz.')  # type: float

    def __init__(self, FxMin=None, FxMax=None, **kwargs):
        """

        Parameters
        ----------
        FxMin : float
        FxMax : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.FxMin = FxMin
        self.FxMax = FxMax
        super(FxBandType, self).__init__(**kwargs)

    def get_array(self, dtype=numpy.float64):
        return numpy.array([self.FxMin, self.FxMax], dtype=dtype)

    @classmethod
    def from_array(cls, array):
        # type: (Union[numpy.ndarray, list, tuple]) -> FxBandType
        if array is None:
            return None
        if isinstance(array, (numpy.ndarray, list, tuple)):
            if len(array) < 2:
                raise ValueError('Expected array to be of length 2, and received {}'.format(array))
            return cls(FxMin=array[0], FxMax=array[1])
        raise ValueError('Expected array to be numpy.ndarray, list, or tuple, got {}'.format(type(array)))


class TOASwathType(Serializable, Arrayable):
    """
    Parameters that describe the time-of-arrival (TOA) swath limits for the signal
    array(s) contained in the product.
    """

    _fields = ('TOAMin', 'TOAMax')
    _required = _fields
    _numeric_format = {fld: FLOAT_FORMAT for fld in _fields}
    # descriptors
    TOAMin = FloatDescriptor(
        'TOAMin', _required, strict=DEFAULT_STRICT,
        docstring=r'Minimum :math:`\Delta TOA` value for any signal vector in '
                  'the product, in seconds.')  # type: float
    TOAMax = FloatDescriptor(
        'TOAMax', _required, strict=DEFAULT_STRICT,
        docstring=r'Maximum :math:`\Delta TOA` value for any signal vector in '
                  'the product, in seconds.')  # type: float

    def __init__(self, TOAMin=None, TOAMax=None, **kwargs):
        """

        Parameters
        ----------
        TOAMin : float
        TOAMax : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.TOAMin = TOAMin
        self.TOAMax = TOAMax
        super(TOASwathType, self).__init__(**kwargs)

    def get_array(self, dtype=numpy.float64):
        return numpy.array([self.TOAMin, self.TOAMax], dtype=dtype)

    @classmethod
    def from_array(cls, array):
        # type: (Union[numpy.ndarray, list, tuple]) -> TOASwathType
        if array is None:
            return None
        if isinstance(array, (numpy.ndarray, list, tuple)):
            if len(array) != 2:
                raise ValueError('Expected array to be of length 2, and received {}'.format(array))
            return cls(TOAMin=array[0], TOAMax=array[1])
        raise ValueError('Expected array to be numpy.ndarray, list, or tuple, got {}'.format(type(array)))


class TropoParametersType(Serializable):
    """
    Parameters used to compute the propagation delay due to the troposphere.
    """

    _fields = ('N0', 'RefHeight')
    _required = _fields
    _numeric_format = {'N0': FLOAT_FORMAT}
    # descriptors
    N0 = FloatDescriptor(
        'N0', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Refractivity value of the troposphere for the imaged scene used '
                  'to form the product (dimensionless). Value at the IARP '
                  'location.')  # type: float
    RefHeight = StringEnumDescriptor(
        'RefHeight', ('IARP', 'ZERO'), _required, strict=DEFAULT_STRICT,
        docstring='Reference Height for the `N0` value.')  # type: str

    def __init__(self, N0=None, RefHeight=None, **kwargs):
        """

        Parameters
        ----------
        N0 : float
        RefHeight : str
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.N0 = N0
        self.RefHeight = RefHeight
        super(TropoParametersType, self).__init__(**kwargs)


class IonoParametersType(Serializable):
    """
    Parameters used to compute propagation effects due to the ionosphere.
    """

    _fields = ('TECV', 'F2Height')
    _required = ('TECV', )
    _numeric_format = {fld: FLOAT_FORMAT for fld in _fields}
    # descriptor
    TECV = FloatDescriptor(
        'TECV', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Total Electron Content (TEC) integrated along TECU the Vertical (V), '
                  r'in units where :math:`1 TECU = 10^{16} e^{-}/m^{2}`')  # type: float
    F2Height = FloatDescriptor(
        'F2Height', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='The F2 height of the ionosphere, in '
                  'meters.')  # type: Union[None, float]

    def __init__(self, TECV=None, F2Height=None, **kwargs):
        """

        Parameters
        ----------
        TECV : float
        F2Height : None|float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.TECV = TECV
        self.F2Height = F2Height
        super(IonoParametersType, self).__init__(**kwargs)


class GlobalType(Serializable):
    """
    The Global type definition.
    """

    _fields = (
        'DomainType', 'SGN', 'Timeline', 'FxBand', 'TOASwath', 'TropoParameters', 'IonoParameters')
    _required = ('DomainType', 'SGN', 'Timeline', 'FxBand', 'TOASwath')
    # descriptors
    DomainType = StringEnumDescriptor(
        'DomainType', ('FX', 'TOA'), _required, strict=DEFAULT_STRICT,
        docstring='Indicates the domain represented by the sample dimension of the '
                  'CPHD signal array(s), where "FX" denotes Transmit Frequency, and '
                  '"TOA" denotes Difference in Time of Arrival')  # type: str
    SGN = IntegerEnumDescriptor(
        'SGN', (-1, 1), _required, strict=DEFAULT_STRICT,
        docstring='Phase SGN applied to compute target signal phase as a function of '
                  r'target :math:`\Delta TOA^{TGT}`. Target phase in cycles. '
                  r'For simple phase model :math:`Phase(fx) = SGN \times fx \times \Delta TOA^{TGT}` '
                  r'In TOA domain, phase of the mainlobe peak '
                  r':math:`Phase(\Delta TOA^{TGT}) = SGN \times fx_C \times \Delta TOA^{TGT}`'
                  '.')  # type: int
    Timeline = SerializableDescriptor(
        'Timeline', TimelineType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that describe the collection times for the data contained '
                  'in the product')  # type: TimelineType
    FxBand = SerializableDescriptor(
        'FxBand', FxBandType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that describe the FX frequency limits for the signal array(s) '
                  'contained in the product.')  # type: FxBandType
    TOASwath = SerializableDescriptor(
        'TOASwath', TOASwathType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that describe the time-of-arrival (TOA) swath limits for the '
                  'signal array(s) contained in the product.')  # type: TOASwathType
    TropoParameters = SerializableDescriptor(
        'TropoParameters', TropoParametersType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters used to compute the propagation delay due to the '
                  'troposphere.')  # type: Union[None, TropoParametersType]
    IonoParameters = SerializableDescriptor(
        'IonoParameters', IonoParametersType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters used to compute propagation effects due to the '
                  'ionosphere.')  # type: Union[None, IonoParametersType]

    def __init__(self, DomainType=None, SGN=None, Timeline=None, FxBand=None, TOASwath=None,
                 TropoParameters=None, IonoParameters=None, **kwargs):
        """

        Parameters
        ----------
        DomainType : str
        SGN : int
        Timeline : TimelineType
        FxBand : FxBandType|numpy.ndarray|list|tuple
        TOASwath : TOASwathType|numpy.ndarray|list|tuple
        TropoParameters : None|TropoParametersType
        IonoParameters : None|IonoParametersType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.DomainType = DomainType
        self.SGN = SGN
        self.Timeline = Timeline
        self.FxBand = FxBand
        self.TOASwath = TOASwath
        self.TropoParameters = TropoParameters
        self.IonoParameters = IonoParameters
        super(GlobalType, self).__init__(**kwargs)




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class EXPLTA_87Type(TREElement):
    def __init__(self, value):
        super(EXPLTA_87Type, self).__init__()
        self.add_field('ANGLE_TO_NORTH', 's', 3, value)
        self.add_field('SQUINT_ANGLE', 's', 3, value)
        self.add_field('MODE', 's', 3, value)
        self.add_field('RESVD001', 's', 16, value)
        self.add_field('GRAZE_ANG', 's', 2, value)
        self.add_field('SLOPE_ANG', 's', 2, value)
        self.add_field('POLAR', 's', 2, value)
        self.add_field('NSAMP', 's', 5, value)
        self.add_field('RESVD002', 's', 1, value)
        self.add_field('SEQ_NUM', 's', 1, value)
        self.add_field('PRIME_ID', 's', 12, value)
        self.add_field('PRIME_BE', 's', 15, value)
        self.add_field('RESVD003', 's', 1, value)
        self.add_field('N_SEC', 's', 2, value)
        self.add_field('IPR', 's', 2, value)
        self.add_field('RESVD004', 's', 2, value)
        self.add_field('RESVD005', 's', 2, value)
        self.add_field('RESVD006', 's', 5, value)
        self.add_field('RESVD007', 's', 8, value)


class EXPLTA_87(TREExtension):
    _tag_value = 'EXPLTA'
    _data_type = EXPLTA_87Type


class EXPLTA_101Type(TREElement):
    def __init__(self, value):
        super(EXPLTA_101Type, self).__init__()
        self.add_field('ANGLE_TO_NORTH', 's', 7, value)
        self.add_field('SQUINT_ANGLE', 's', 7, value)
        self.add_field('MODE', 's', 3, value)
        self.add_field('RESVD001', 's', 16, value)
        self.add_field('GRAZE_ANG', 's', 5, value)
        self.add_field('SLOPE_ANG', 's', 5, value)
        self.add_field('POLAR', 's', 2, value)
        self.add_field('NSAMP', 's', 5, value)
        self.add_field('RESVD002', 's', 1, value)
        self.add_field('SEQ_NUM', 's', 1, value)
        self.add_field('PRIME_ID', 's', 12, value)
        self.add_field('PRIME_BE', 's', 15, value)
        self.add_field('RESVD003', 's', 1, value)
        self.add_field('N_SEC', 's', 2, value)
        self.add_field('IPR', 's', 2, value)


class EXPLTA_101(TREExtension):
    _tag_value = 'EXPLTA'
    _data_type = EXPLTA_101Type


class EXPLTA(TREExtension):
    _tag_value = 'EXPLTA'

    def __init__(self):
        raise ValueError(
            'Not to be implemented directly. '
            'Use of one EXPLTA_87 or  EXPLTA_101')

    @classmethod
    def from_bytes(cls, value, start):
        """

        Parameters
        ----------
        value : bytes
        start : int

        Returns
        -------
        EXPLTA_87|EXPLTA_101
        """

        tag_value = value[start:start+6].decode('utf-8').strip()
        if tag_value != cls._tag_value:
            raise ValueError('tag value must be {}. Got {}'.format(cls._tag_value, tag_value))

        lng = int(value[start+6:start+11])
        if lng == 87:
            return EXPLTA_87.from_bytes(value, start)
        elif lng == 101:
            return EXPLTA_101.from_bytes(value, start)
        else:
            raise ValueError('the data must be length 87 or 101. Got {}'.format(lng))



"""
The Global type definition for CPHD 0.3.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from typing import Union, List

import numpy

from sarpy.io.phase_history.cphd1_elements.base import DEFAULT_STRICT, FLOAT_FORMAT
from sarpy.io.complex.sicd_elements.base import SerializableCPArrayDescriptor, SerializableCPArray
from sarpy.io.complex.sicd_elements.blocks import LatLonHAECornerRestrictionType, Poly2DType
from sarpy.io.complex.sicd_elements.RadarCollection import ReferencePointType, XDirectionType, \
    YDirectionType
from sarpy.io.xml.base import Serializable
from sarpy.io.xml.descriptors import FloatDescriptor, DateTimeDescriptor, \
    StringEnumDescriptor, IntegerDescriptor, IntegerEnumDescriptor, \
    SerializableDescriptor


class DwellTimeType(Serializable):
    """
    The dwell time object.
    """
    _fields = ('DwellTimePoly', 'CODTimePoly')
    _required = _fields
    # descriptors
    DwellTimePoly = SerializableDescriptor(
        'DwellTimePoly', Poly2DType, _required, strict=DEFAULT_STRICT,
        docstring='The dwell time polynomial.')  # type: Poly2DType
    CODTimePoly = SerializableDescriptor(
        'CODTimePoly', Poly2DType, _required, strict=DEFAULT_STRICT,
        docstring='The cod time polynomial.')  # type: Poly2DType

    def __init__(self, DwellTimePoly=None, CODTimePoly=None, **kwargs):
        """

        Parameters
        ----------
        DwellTimePoly : Poly2DType|numpy.ndarray|list|tuple
        CODTimePoly : Poly2DType|numpy.ndarray|list|tuple
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.DwellTimePoly = DwellTimePoly
        self.CODTimePoly = CODTimePoly
        super(DwellTimeType, self).__init__(**kwargs)


class PlaneType(Serializable):
    """
    The reference plane.
    """

    _fields = ('RefPt', 'XDir', 'YDir', 'DwellTime')
    _required = ('RefPt', 'XDir', 'YDir')
    # other class variable
    # descriptors
    RefPt = SerializableDescriptor(
        'RefPt', ReferencePointType, _required, strict=DEFAULT_STRICT,
        docstring='The reference point.')  # type: ReferencePointType
    XDir = SerializableDescriptor(
        'XDir', XDirectionType, _required, strict=DEFAULT_STRICT,
        docstring='The X direction collection plane parameters.')  # type: XDirectionType
    YDir = SerializableDescriptor(
        'YDir', YDirectionType, _required, strict=DEFAULT_STRICT,
        docstring='The Y direction collection plane parameters.')  # type: YDirectionType
    DwellTime = SerializableDescriptor(
        'DwellTime', DwellTimeType, _required, strict=DEFAULT_STRICT,
        docstring='The dwell time parameters.')  # type: DwellTimeType

    def __init__(self, RefPt=None, XDir=None, YDir=None, DwellTime=None, **kwargs):
        """

        Parameters
        ----------
        RefPt : ReferencePointType
        XDir : XDirectionType
        YDir : YDirectionType
        DwellTime : DwellTimeType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.RefPt = RefPt
        self.XDir = XDir
        self.YDir = YDir
        self.DwellTime = DwellTime
        super(PlaneType, self).__init__(**kwargs)


class ImageAreaType(Serializable):
    """
    The collection area.
    """

    _fields = ('Corner', 'Plane')
    _required = ('Corner', )
    _collections_tags = {
        'Corner': {'array': True, 'child_tag': 'ACP'}, }
    # descriptors
    Corner = SerializableCPArrayDescriptor(
        'Corner', LatLonHAECornerRestrictionType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='The collection area corner point definition array.'
    )  # type: Union[SerializableCPArray, List[LatLonHAECornerRestrictionType]]
    Plane = SerializableDescriptor(
        'Plane', PlaneType, _required, strict=DEFAULT_STRICT,
        docstring='A rectangular area in a geo-located display plane.')  # type: PlaneType

    def __init__(self, Corner=None, Plane=None, **kwargs):
        """

        Parameters
        ----------
        Corner : SerializableCPArray|List[LatLonHAECornerRestrictionType]|numpy.ndarray|list|tuple
        Plane : PlaneType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Corner = Corner
        self.Plane = Plane
        super(ImageAreaType, self).__init__(**kwargs)


class GlobalType(Serializable):
    """
    The Global type definition.
    """

    _fields = (
        'DomainType', 'PhaseSGN', 'RefFreqIndex', 'CollectStart',
        'CollectDuration', 'TxTime1', 'TxTime2', 'ImageArea')
    _required = (
        'DomainType', 'PhaseSGN', 'CollectStart', 'CollectDuration',
        'TxTime1', 'TxTime2', 'ImageArea')
    _numeric_format = {
        'CollectDuration': FLOAT_FORMAT, 'TxTime1': FLOAT_FORMAT, 'TxTime2': FLOAT_FORMAT}
    # descriptors
    DomainType = StringEnumDescriptor(
        'DomainType', ('FX', 'TOA'), _required, strict=DEFAULT_STRICT,
        docstring='Indicates the domain represented by the sample dimension of the '
                  'CPHD signal array(s), where "FX" denotes Transmit Frequency, and '
                  '"TOA" denotes Difference in Time of Arrival')  # type: str
    PhaseSGN = IntegerEnumDescriptor(
        'PhaseSGN', (-1, 1), _required, strict=DEFAULT_STRICT,
        docstring='Phase SGN applied to compute target signal phase as a function of '
                  r'target :math:`\Delta TOA^{TGT}`. Target phase in cycles. '
                  r'For simple phase model :math:`Phase(fx) = SGN \times fx \times \Delta TOA^{TGT}` '
                  r'In TOA domain, phase of the mainlobe peak '
                  r':math:`Phase(\Delta TOA^{TGT}) = SGN \times fx_C \times \Delta TOA^{TGT}`'
                  '.')  # type: int
    RefFreqIndex = IntegerDescriptor(
        'RefFreqIndex', _required, strict=DEFAULT_STRICT,
        docstring='Indicates if the RF frequency values are expressed as offsets from '
                  'a reference frequency (RefFreq).')  # type: Union[None, int]
    CollectStart = DateTimeDescriptor(
        'CollectStart', _required, strict=DEFAULT_STRICT, numpy_datetime_units='us',
        docstring='Collection Start date and time (UTC). Time reference used for times '
                  'measured from collection start (i.e. slow time t = 0). For bistatic '
                  'collections, the time is the transmit platform collection '
                  'start time. The default display precision is microseconds, but this '
                  'does not that accuracy in value.')  # type: numpy.datetime64
    CollectDuration = FloatDescriptor(
        'CollectDuration', _required, strict=DEFAULT_STRICT,
        docstring='The duration of the collection, in seconds.')  # type: float
    TxTime1 = FloatDescriptor(
        'TxTime1', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Earliest TxTime value for any signal vector in the product. '
                  'Time relative to Collection Start in seconds.')  # type: float
    TxTime2 = FloatDescriptor(
        'TxTime2', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Latest TxTime value for any signal vector in the product. '
                  'Time relative to Collection Start in seconds.')  # type: float
    ImageArea = SerializableDescriptor(
        'ImageArea', ImageAreaType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters describing the ground area covered by this '
                  'product.')  # type: ImageAreaType

    def __init__(self, DomainType=None, PhaseSGN=None, RefFreqIndex=None, CollectStart=None,
                 CollectDuration=None, TxTime1=None, TxTime2=None, ImageArea=None, **kwargs):
        """

        Parameters
        ----------
        DomainType : str
        PhaseSGN : int
        RefFreqIndex : None|int
        CollectStart : numpy.datetime64|datetime.datetime|str
        CollectDuration : float
        TxTime1 : float
        TxTime2 : float
        ImageArea : ImageAreaType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.DomainType = DomainType
        self.PhaseSGN = PhaseSGN
        self.RefFreqIndex = RefFreqIndex
        self.CollectStart = CollectStart
        self.CollectDuration = CollectDuration
        self.TxTime1 = TxTime1
        self.TxTime2 = TxTime2
        self.ImageArea = ImageArea
        super(GlobalType, self).__init__(**kwargs)



"""
Definition for the SubCollectionInfo NGA modified RDE/AFRL labeling object
"""

__classification__ = "UNCLASSIFIED"
__authors__ = "Thomas McCullough"


from typing import Optional

from sarpy.io.xml.base import Serializable
from sarpy.io.xml.descriptors import SerializableDescriptor, StringDescriptor

from .base import DEFAULT_STRICT
from .blocks import DateTimeRangeType, LatLonEleType


class SubCollectionInfoType(Serializable):
    _fields = ('Name', 'SiteName', 'SiteNumber', 'SceneNumber', 'Description',
               'Duration', 'SiteCenterLocation', 'SceneContentDescription',
               'SiteBackgroundType')
    _required = ('Name', 'SiteCenterLocation', 'SceneContentDescription')
    # descriptors
    Name = StringDescriptor(
        'Name', _required,
        docstring="Name of the subcollection.")  # type: str
    SiteName = StringDescriptor(
        'SiteName', _required,
        docstring="Name of the subcollection site location.")  # type: Optional[str]
    SiteNumber = StringDescriptor(
        'SiteNumber', _required,
        docstring="Site number of the subcollection.")  # type: Optional[str]
    SceneNumber = StringDescriptor(
        'SceneNumber', _required,
        docstring="Scene number of the subcollection.")  # type: Optional[str]
    Description = StringDescriptor(
        'Description', _required,
        docstring="Description of the subcollection (e.g., Main array).")  # type: Optional[str]
    Duration = SerializableDescriptor(
        'Duration', DateTimeRangeType, _required, strict=DEFAULT_STRICT,
        docstring="Begin and end dates of the subcollection.")  # type: Optional[DateTimeRangeType]
    SiteCenterLocation = SerializableDescriptor(
        'SiteCenterLocation', LatLonEleType, _required, strict=DEFAULT_STRICT,
        docstring="Location of the center of the collection site.")  # type: LatLonEleType
    SceneContentDescription = StringDescriptor(
        'SceneContentDescription', _required, default_value="",
        docstring="Description of the general scene contents.")  # type: str
    SiteBackgroundType = StringDescriptor(
        'SiteBackgroundType', _required,
        docstring="Description of the background.")  # type: Optional[str]

    def __init__(self, Name=None, SiteName=None, SiteNumber=None,
                 SceneNumber=None, Description=None, Duration=None,
                 SiteCenterLocation=None, SceneContentDescription=None,
                 SiteBackgroundType=None, **kwargs):
        """
        Parameters
        ----------
        Name : None|str
        SiteName : None|str
        SiteNumber : None|str
        SceneNumber : None|str
        Description : NOne|str
        Duration : None|DateTimeRangeType|list|tuple
        SiteCenterLocation : LatLonEleType|numpy.ndarray|list|tuple
        SceneContentDescription : None|str
        SiteBackgroundType : None|str
        kwargs
            Other keyword arguments
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Name = Name
        self.SiteName = SiteName
        self.SiteNumber = SiteNumber
        self.SceneNumber = SceneNumber
        self.Description = Description
        self.Duration = Duration
        self.SiteCenterLocation = SiteCenterLocation
        self.SceneContentDescription = SceneContentDescription
        self.SiteBackgroundType = SiteBackgroundType
        super(SubCollectionInfoType, self).__init__(**kwargs)




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class CSSHPAType(TREElement):
    def __init__(self, value):
        super(CSSHPAType, self).__init__()
        self.add_field('SHAPE_USE', 's', 25, value)
        self.add_field('SHAPE_CLASS', 's', 10, value)
        if self.SHAPE_USE.strip() == 'CLOUD_SHAPES':
            self.add_field('CC_SOURCE', 's', 18, value)
        self.add_field('SHAPE1_NAME', 's', 3, value)
        self.add_field('SHAPE1_START', 's', 6, value)
        self.add_field('SHAPE2_NAME', 's', 3, value)
        self.add_field('SHAPE2_START', 's', 6, value)
        self.add_field('SHAPE3_NAME', 's', 3, value)
        self.add_field('SHAPE3_START', 's', 6, value)


class CSSHPA(TREExtension):
    _tag_value = 'CSSHPA'
    _data_type = CSSHPAType




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class GEOLOBType(TREElement):
    def __init__(self, value):
        super(GEOLOBType, self).__init__()
        self.add_field('ARV', 's', 9, value)
        self.add_field('BRV', 's', 9, value)
        self.add_field('LSO', 's', 15, value)
        self.add_field('PSO', 's', 15, value)


class GEOLOB(TREExtension):
    _tag_value = 'GEOLOB'
    _data_type = GEOLOBType



"""
The DigitalElevationDataType definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from typing import Union, List

import numpy

from sarpy.io.xml.base import Serializable
from sarpy.io.xml.descriptors import SerializableDescriptor, IntegerDescriptor, \
    FloatDescriptor, FloatListDescriptor, StringEnumDescriptor

from .base import DEFAULT_STRICT, FLOAT_FORMAT
from .blocks import LatLonType


class GeographicCoordinatesType(Serializable):
    """
    Describes the Local Geographic Coordinate system linking row/column to the absolute
    geographic coordinate (lat/lon).
    """

    _fields = ('LongitudeDensity', 'LatitudeDensity', 'ReferenceOrigin')
    _required = ('LongitudeDensity', 'LatitudeDensity', 'ReferenceOrigin')
    _numeric_format = {'LongitudeDensity': FLOAT_FORMAT, 'LatitudeDensity': FLOAT_FORMAT}
    # Descriptor
    LongitudeDensity = FloatDescriptor(
        'LongitudeDensity', _required, strict=DEFAULT_STRICT,
        docstring='Pixel ground spacing in E/W direction that is the number of pixels '
                  'or element intervals in 360 degrees.')  # type: float
    LatitudeDensity = FloatDescriptor(
        'LatitudeDensity', _required, strict=DEFAULT_STRICT,
        docstring='Pixel ground spacing in N/S direction that is the number of pixels '
                  'or element intervals in 360 degrees.')  # type: float
    ReferenceOrigin = SerializableDescriptor(
        'ReferenceOrigin', LatLonType, _required, strict=DEFAULT_STRICT,
        docstring='Northwest corner Latitude/Longitude - product NW corner')  # type: LatLonType

    def __init__(self, LongitudeDensity=None, LatitudeDensity=None, ReferenceOrigin=None, **kwargs):
        """

        Parameters
        ----------
        LongitudeDensity : float
        LatitudeDensity : float
        ReferenceOrigin : LatLonType|numpy.ndarray|list|tuple
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        self.LongitudeDensity = LongitudeDensity
        self.LatitudeDensity = LatitudeDensity
        self.ReferenceOrigin = ReferenceOrigin
        super(GeographicCoordinatesType, self).__init__(**kwargs)


class GeopositioningType(Serializable):
    """
    Describes the absolute coordinate system to which the data is referenced.
    """
    _fields = (
        'CoordinateSystemType', 'GeodeticDatum', 'ReferenceEllipsoid',
        'VerticalDatum', 'SoundingDatum', 'FalseOrigin', 'UTMGridZoneNumber')
    _required = (
        'CoordinateSystemType', 'GeodeticDatum', 'ReferenceEllipsoid',
        'VerticalDatum', 'SoundingDatum', 'FalseOrigin')
    # Descriptor
    CoordinateSystemType = StringEnumDescriptor(
        'CoordinateSystemType', ('GCS', 'UTM'), _required, strict=DEFAULT_STRICT,
        docstring='')  # type: str
    GeodeticDatum = StringEnumDescriptor(
        'GeodeticDatum', ('World Geodetic System 1984', ), _required, strict=DEFAULT_STRICT,
        default_value='World Geodetic System 1984',
        docstring='')  # type: str
    ReferenceEllipsoid = StringEnumDescriptor(
        'ReferenceEllipsoid', ('World Geodetic System 1984', ), _required, strict=DEFAULT_STRICT,
        default_value='World Geodetic System 1984',
        docstring='')  # type: str
    VerticalDatum = StringEnumDescriptor(
        'VerticalDatum', ('Mean Sea Level', ), _required, strict=DEFAULT_STRICT,
        default_value='Mean Sea Level',
        docstring='')  # type: str
    SoundingDatum = StringEnumDescriptor(
        'SoundingDatum', ('Mean Sea Level', ), _required, strict=DEFAULT_STRICT,
        default_value='Mean Sea Level',
        docstring='')  # type: str
    FalseOrigin = IntegerDescriptor(
        'FalseOrigin', _required, strict=DEFAULT_STRICT,
        docstring='Z values false origin.')  # type: int
    UTMGridZoneNumber = IntegerDescriptor(
        'UTMGridZoneNumber', _required, strict=DEFAULT_STRICT,
        docstring='Grid zone number, required for UTM, not include for GCS. '
                  '**Values -** `+001` to `+060` (northern hemisphere) and `-001` to `-060` '
                  '(southern hemisphere)')  # type: int

    def __init__(self, CoordinateSystemType=None, GeodeticDatum='World Geodetic System 1984',
                 ReferenceEllipsoid='World Geodetic System 1984', VerticalDatum='Mean Sea Level',
                 SoundingDatum='Mean Sea Level', FalseOrigin=None, UTMGridZoneNumber=None, **kwargs):
        """

        Parameters
        ----------
        CoordinateSystemType : str
        GeodeticDatum : str
        ReferenceEllipsoid : str
        VerticalDatum : str
        SoundingDatum : str
        FalseOrigin : int
        UTMGridZoneNumber : None|int
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        self.CoordinateSystemType = CoordinateSystemType
        self.GeodeticDatum = GeodeticDatum
        self.ReferenceEllipsoid = ReferenceEllipsoid
        self.VerticalDatum = VerticalDatum
        self.SoundingDatum = SoundingDatum
        self.FalseOrigin = FalseOrigin
        self.UTMGridZoneNumber = UTMGridZoneNumber
        super(GeopositioningType, self).__init__(**kwargs)


class AccuracyType(Serializable):
    """
    The accuracy estimate.
    """

    _fields = ('Horizontals', 'Verticals')
    _required = ('Horizontals', 'Verticals')
    _collections_tags = {
        'Horizontals': {'array': False, 'child_tag': 'Horizontal'},
        'Verticals': {'array': False, 'child_tag': 'Vertical'}}
    _numeric_format = {key: FLOAT_FORMAT for key in _fields}
    # Descriptor
    Horizontals = FloatListDescriptor(
        'Horizontals', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: List[float]
    Verticals = FloatListDescriptor(
        'Verticals', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: List[float]

    def __init__(self, Horizontals=None, Verticals=None, **kwargs):
        """

        Parameters
        ----------
        Horizontals : List[float]
        Verticals : List[float]
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        self.Horizontals = Horizontals
        self.Verticals = Verticals
        super(AccuracyType, self).__init__(**kwargs)


class PositionalAccuracyType(Serializable):
    """
    Describes the horizontal and vertical point and regional information for the DED.
    """

    _fields = ('NumRegions', 'AbsoluteAccuracy', 'PointToPointAccuracy')
    _required = ('NumRegions', 'AbsoluteAccuracy', 'PointToPointAccuracy')
    # Descriptor
    NumRegions = IntegerDescriptor(
        'NumRegions', _required, strict=DEFAULT_STRICT,
        docstring='Number of positional accuracy regions.')  # type: int
    AbsoluteAccuracy = SerializableDescriptor(
        'AbsoluteAccuracy', AccuracyType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: AccuracyType
    PointToPointAccuracy = SerializableDescriptor(
        'PointToPointAccuracy', AccuracyType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: AccuracyType

    def __init__(self, NumRegions=None, AbsoluteAccuracy=None, PointToPointAccuracy=None, **kwargs):
        """

        Parameters
        ----------
        NumRegions : int
        AbsoluteAccuracy : AccuracyType
        PointToPointAccuracy : AccuracyType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.NumRegions = NumRegions
        self.AbsoluteAccuracy = AbsoluteAccuracy
        self.PointToPointAccuracy = PointToPointAccuracy
        super(PositionalAccuracyType, self).__init__(**kwargs)


class DigitalElevationDataType(Serializable):
    """
    This describes any Digital Elevation Data included with the SIDD product.
    """

    _fields = ('GeographicCoordinates', 'Geopositioning', 'PositionalAccuracy', 'NullValue')
    _required = ('GeographicCoordinates', 'Geopositioning', 'PositionalAccuracy')
    # Descriptor
    GeographicCoordinates = SerializableDescriptor(
        'GeographicCoordinates', GeographicCoordinatesType, _required, strict=DEFAULT_STRICT,
        docstring='Describes the Local Geographic Coordinate system linking row/column to the '
                  'absolute geographic coordinate (lat/lon)')  # type: GeographicCoordinatesType
    Geopositioning = SerializableDescriptor(
        'Geopositioning', GeopositioningType, _required, strict=DEFAULT_STRICT,
        docstring='Describes the absolute coordinate system to which the data is '
                  'referenced.')  # type: GeopositioningType
    PositionalAccuracy = SerializableDescriptor(
        'PositionalAccuracy', PositionalAccuracyType, _required, strict=DEFAULT_STRICT,
        docstring='Describes the horizontal and vertical point and regional information '
                  'for the DED.')  # type: PositionalAccuracyType
    NullValue = IntegerDescriptor(
        'NullValue', _required, strict=DEFAULT_STRICT,
        docstring='The value in the DEM corresponding to `No Value`.')  # type: Union[None, int]

    def __init__(self, GeographicCoordinates=None, Geopositioning=None, PositionalAccuracy=None,
                 NullValue=None, **kwargs):
        """

        Parameters
        ----------
        GeographicCoordinates : GeographicCoordinatesType
        Geopositioning : GeopositioningType
        PositionalAccuracy : PositionalAccuracyType
        NullValue : int
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.GeographicCoordinates = GeographicCoordinates
        self.Geopositioning = Geopositioning
        self.PositionalAccuracy = PositionalAccuracy
        self.NullValue = NullValue
        super(DigitalElevationDataType, self).__init__(**kwargs)



"""
Multipurpose basic SIDD elements
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


from typing import Union, Optional

import numpy

from sarpy.io.xml.base import Serializable, Arrayable
from sarpy.io.xml.descriptors import FloatDescriptor, StringEnumDescriptor

from .base import DEFAULT_STRICT, FLOAT_FORMAT
import sarpy.io.complex.sicd_elements.Radiometric as SicdRadiometric

# Reuse SIDD 2.0 types and make them available here
from sarpy.io.product.sidd2_elements.blocks import (
    ErrorStatisticsType,
    FilterType,
    FilterBankType,
    MatchInfoType,
    NewLookupTableType,
    PredefinedFilterType,
    PredefinedLookupType,
    Poly2DType,
    RadarModeType,
    RangeAzimuthType,
    ReferencePointType,
    RowColDoubleType,
    XYZPolyType,
    XYZType,
)

__REUSED__ = (  # to avoid unused imports
    ErrorStatisticsType,
    FilterType,
    FilterBankType,
    MatchInfoType,
    NewLookupTableType,
    PredefinedFilterType,
    PredefinedLookupType,
    Poly2DType,
    RadarModeType,
    RangeAzimuthType,
    ReferencePointType,
    RowColDoubleType,
    XYZPolyType,
    XYZType,
)

_len2_array_text = 'Expected array to be of length 2,\n\t' \
                   'and received `{}`'
_array_type_text = 'Expected array to be numpy.ndarray, list, or tuple,\n\t' \
                   'got `{}`'


############
# the SICommon namespace elements

class AngleZeroToExclusive360MagnitudeType(Serializable, Arrayable):
    """
    Represents a magnitude and angle.
    """

    _fields = ('Angle', 'Magnitude')
    _required = ('Angle', 'Magnitude')
    _numeric_format = {key: FLOAT_FORMAT for key in _fields}
    _child_xml_ns_key = {'Angle': 'sicommon', 'Magnitude': 'sicommon'}
    # Descriptor
    Angle = FloatDescriptor(
        'Angle', _required, strict=DEFAULT_STRICT, bounds=(0.0, 360),
        docstring='The angle.')  # type: float
    Magnitude = FloatDescriptor(
        'Magnitude', _required, strict=DEFAULT_STRICT, bounds=(0.0, None),
        docstring='The magnitude.')  # type: float

    def __init__(self, Angle=None, Magnitude=None, **kwargs):
        """

        Parameters
        ----------
        Angle : float
        Magnitude : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Angle = Angle
        self.Magnitude = Magnitude
        super(AngleZeroToExclusive360MagnitudeType, self).__init__(**kwargs)

    def get_array(self, dtype=numpy.float64):
        """
        Gets an array representation of the class instance.

        Parameters
        ----------
        dtype : str|numpy.dtype|numpy.number
            numpy data type of the return

        Returns
        -------
        numpy.ndarray
            array of the form [Angle, Magnitude]
        """

        return numpy.array([self.Angle, self.Magnitude], dtype=dtype)

    @classmethod
    def from_array(cls, array):
        """
        Create from an array type entry.

        Parameters
        ----------
        array: numpy.ndarray|list|tuple
            assumed [Angle, Magnitude]

        Returns
        -------
        AngleZeroToExclusive360MagnitudeType
        """

        if array is None:
            return None
        if isinstance(array, (numpy.ndarray, list, tuple)):
            if len(array) < 2:
                raise ValueError(_len2_array_text.format(array))
            return cls(Angle=array[0], Magnitude=array[1])
        raise ValueError(_array_type_text.format(type(array)))


# SIDD 3.0 Radiometric is the same as SICD, but with an added field
class RadiometricType(SicdRadiometric.RadiometricType):
    _fields = ('NoiseLevel', 'RCSSFPoly', 'SigmaZeroSFPoly', 'BetaZeroSFPoly', 'SigmaZeroSFIncidenceMap',
               'GammaZeroSFPoly')
    _child_xml_ns_key = {
        'NoiseLevel': 'sicommon', 'RCSSFPoly': 'sicommon', 'SigmaZeroSFPoly': 'sicommon',
        'BetaZeroSFPoly': 'sicommon', 'SigmaZeroSFIncidenceMap': 'sicommon', 'GammaZeroSFPoly': 'sicommon'}

    _SIGMA_ZERO_SF_INCIDENCE_MAP_VALUES = ('APPLIED', 'NOT_APPLIED')
    # descriptors
    SigmaZeroSFIncidenceMap = StringEnumDescriptor(
        'SigmaZeroSFIncidenceMap', _SIGMA_ZERO_SF_INCIDENCE_MAP_VALUES,
        SicdRadiometric.RadiometricType._required, strict=DEFAULT_STRICT,
        docstring='Allowed Values: “APPLIED” or “NOT_APPLIED”')  # type: str

    def __init__(
            self,
            NoiseLevel: Optional[SicdRadiometric.NoiseLevelType_] = None,
            RCSSFPoly: Union[None, Poly2DType, numpy.ndarray, list, tuple] = None,
            SigmaZeroSFPoly: Union[None, Poly2DType, numpy.ndarray, list, tuple] = None,
            BetaZeroSFPoly: Union[None, Poly2DType, numpy.ndarray, list, tuple] = None,
            GammaZeroSFPoly: Union[None, Poly2DType, numpy.ndarray, list, tuple] = None,
            SigmaZeroSFIncidenceMap: str = None,
            **kwargs):
        """

        Parameters
        ----------
        NoiseLevel : NoiseLevelType_
        RCSSFPoly : Poly2DType|numpy.ndarray|list|tuple
        SigmaZeroSFPoly : Poly2DType|numpy.ndarray|list|tuple
        BetaZeroSFPoly : Poly2DType|numpy.ndarray|list|tuple
        GammaZeroSFPoly : Poly2DType|numpy.ndarray|list|tuple
        SigmaZeroSFIncidenceMap : str
        kwargs
        """
        self.SigmaZeroSFIncidenceMap = SigmaZeroSFIncidenceMap
        super().__init__(NoiseLevel=NoiseLevel,
                         RCSSFPoly=RCSSFPoly,
                         SigmaZeroSFPoly=SigmaZeroSFPoly,
                         BetaZeroSFPoly=BetaZeroSFPoly,
                         GammaZeroSFPoly=GammaZeroSFPoly,
                         **kwargs)


# The end of the SICommon namespace
#####################



"""
The methods for computing a color sub-aperture image for SICD type images.

As noted in the CSICalculator class, the full resolution data along the split dimension
is required, so sub-sampling along the split dimension does not decrease the amount of
data which must be fetched and/or processing which must be performed.

Examples
--------
.. code-block:: python

    from matplotlib import pyplot
    from sarpy.io.complex.converter import open_complex
    from sarpy.processing.sicd.csi import CSICalculator
    from sarpy.visualization.remap import Density

    # open a sicd type file
    reader = open_complex("<file name>")
    # see the sizes of all image segments
    print(reader.get_data_size_as_tuple())

    # construct the csi performer instance
    # make sure to set the index and dimension as appropriate
    csi_calculator = CSICalculator(reader, dimension=0, index=0)
    # see the size for this particular image element
    # this is identical to the data size from the reader at index
    print(csi_calculator.data_size)

    # set a different index or change the dimension
    # csi_calculator.index = 2
    # csi_calculator.dimension = 1

    # calculate the csi for an image segment
    csi_data = csi_calculator[300:500, 200:600]

    # create our remap function
    density = Density()

    # let's view this csi image using matplotlib
    fig, axs = pyplot.subplots(nrows=1, ncols=1)
    axs.imshow(density(csi_data), aspect='equal')
    pyplot.show()
"""

__classification__ = "UNCLASSIFIED"
__author__ = 'Thomas McCullough'

from typing import Union, Tuple, Optional, Sequence

import numpy

from sarpy.processing.sicd.fft_base import FFTCalculator, fft, ifft, fftshift
from sarpy.io.complex.base import SICDTypeReader
from sarpy.io.complex.utils import get_fetch_block_size
from sarpy.io.general.slice_parsing import get_slice_result_size


def filter_map_construction(siz: Union[int, float]) -> numpy.ndarray:
    """
    Provides the RGB filter array for sub-aperture processing.

    Parameters
    ----------
    siz : int|float
        the size of the colormap

    Returns
    -------
    numpy.ndarray
        the `siz x 3` colormap array
    """

    if siz < 1:
        raise ValueError('Cannot create the filter map with fewer than 4 elements.')

    siz = int(round(siz))
    basic_size = int(numpy.ceil(0.25*siz))

    # create trapezoidal stack
    trapezoid = numpy.hstack(
        (numpy.arange(1, basic_size+1, dtype=numpy.int32),
         numpy.full((basic_size-1, ), basic_size, dtype=numpy.int32),
         numpy.arange(basic_size, 0, -1, dtype=numpy.int32)))/float(basic_size)
    out = numpy.zeros((siz, 3), dtype=numpy.float64)
    # create red, green, blue indices
    green_inds = int(round(0.5*(siz - trapezoid.size))) + numpy.arange(trapezoid.size)
    red_inds = ((green_inds + basic_size) % siz)
    blue_inds = ((green_inds - basic_size) % siz)
    # populate our array
    out[red_inds, 0] = trapezoid
    out[green_inds, 1] = trapezoid
    out[blue_inds, 2] = trapezoid
    return out


def csi_array(
        array: numpy.ndarray,
        dimension: int = 0,
        platform_direction: str = 'R',
        fill: Union[int, float] = 1,
        filter_map: Optional[numpy.ndarray] = None) -> numpy.ndarray:
    """
    Creates a color subaperture array from a complex array.

    .. Note: this ignores any potential sign issues for the fft and ifft, because
        the results would be identical - fft followed by ifft versus ifft followed
        by fft.

    Parameters
    ----------
    array : numpy.ndarray
        The complex valued SAR data, assumed to be in the "image" domain.
        Required to be two-dimensional.
    dimension : int
        The dimension over which to split the sub-aperture.
    platform_direction : str
        The (case insensitive) platform direction, required to be one of `('R', 'L')`.
    fill : float
        The fill factor. This will be ignored if `filter_map` is provided.
    filter_map : None|numpy.ndarray
        The RGB filter mapping. This is assumed constructed using :func:`filter_map_construction`.

    Returns
    -------
    numpy.ndarray
    """

    if not (isinstance(array, numpy.ndarray) and len(array.shape) == 2 and numpy.iscomplexobj(array)):
        raise ValueError('array must be a two-dimensional numpy array of complex dtype')

    dimension = int(dimension)
    if dimension not in [0, 1]:
        raise ValueError('dimension must be 0 or 1, got {}'.format(dimension))
    if dimension == 0:
        array = array.T

    pdir_func = platform_direction.upper()[0]
    if pdir_func not in ['R', 'L']:
        raise ValueError('It is expected that pdir is one of "R" or "L". Got {}'.format(platform_direction))

    # get our filter construction data
    if filter_map is None:
        fill = max(1.0, float(fill))
        filter_map = filter_map_construction(array.shape[1]/fill)
    if not (isinstance(filter_map, numpy.ndarray) and
            filter_map.dtype.name in ['float32', 'float64'] and
            filter_map.ndim == 2 and filter_map.shape[1] == 3):
        raise ValueError('filter_map must be a N x 3 numpy array of float dtype.')

    # move to phase history domain
    ph_indices = int(numpy.floor(0.5*(array.shape[1] - filter_map.shape[0]))) + \
                 numpy.arange(filter_map.shape[0], dtype=numpy.int32)
    ph0 = fftshift(ifft(numpy.asarray(array, numpy.complex128), axis=1), axes=1)[:, ph_indices]
    # construct the filtered workspace
    # NB: processing is more efficient with color band in the first dimension
    ph0_RGB = numpy.zeros((3, array.shape[0], filter_map.shape[0]), dtype=numpy.complex128)
    for i in range(3):
        ph0_RGB[i, :, :] = ph0*filter_map[:, i]
    del ph0

    # Shift phase history to avoid having zeropad in middle of filter, to alleviate
    # the purple sidelobe artifact.
    filter_shift = int(numpy.ceil(0.25*filter_map.shape[0]))
    ph0_RGB[0, :] = numpy.roll(ph0_RGB[0, :], -filter_shift)
    ph0_RGB[2, :] = numpy.roll(ph0_RGB[2, :], filter_shift)
    # NB: the green band is already centered

    # FFT back to the image domain
    im0_RGB = fft(fftshift(ph0_RGB, axes=2), n=array.shape[1], axis=2)
    del ph0_RGB

    # Replace the intensity with the original image intensity to main full resolution
    # (in intensity, but not in color).
    scale_factor = numpy.abs(array)/numpy.abs(im0_RGB).max(axis=0)
    im0_RGB = numpy.abs(im0_RGB)*scale_factor

    # reorient image so that the color segment is in the final dimension
    if dimension == 0:
        im0_RGB = im0_RGB.transpose([2, 1, 0])
    else:
        im0_RGB = im0_RGB.transpose([1, 2, 0])
    if pdir_func == 'R':
        # reverse the color band order
        im0_RGB = im0_RGB[:, :, ::-1]
    return im0_RGB


class CSICalculator(FFTCalculator):
    """
    Class for creating color sub-aperture image from a reader instance.

    It is important to note that full resolution is required for processing along
    the split dimension, so sub-sampling along the split dimension does not decrease
    the amount of data which must be fetched.
    """

    def __init__(
            self,
            reader: Union[str, SICDTypeReader],
            dimension: int = 0,
            index: int = 0,
            block_size: Union[None, int, float] = 50):
        """

        Parameters
        ----------
        reader : str|SICDTypeReader
            Input file path or reader object, which must be of sicd type.
        dimension : int
            The dimension over which to split the sub-aperture.
        index : int
            The sicd index to use.
        block_size : None|int|float
            The approximate processing block size to fetch, given in MB.
        """

        super(CSICalculator, self).__init__(
            reader, dimension=dimension, index=index, block_size=block_size)

    def get_fetch_block_size(
            self,
            start_element: int,
            stop_element: int) -> int:
        """
        Gets the fetch block size for the given full resolution section.
        This assumes that the fetched data will be 24 bytes per pixel, in
        accordance with 3-band complex64 data.

        Parameters
        ----------
        start_element : int
        stop_element : int

        Returns
        -------
        int
        """

        return get_fetch_block_size(start_element, stop_element, self.block_size_in_bytes, bands=3)

    def _full_row_resolution(
            self,
            row_range: Union[slice, Tuple[int, int, int]],
            col_range: Union[slice, Tuple[int, int, int]],
            filter_map: Optional[numpy.ndarray] = None) -> numpy.ndarray:
        data = super(CSICalculator, self)._full_row_resolution(row_range, col_range)
        return csi_array(
            data, dimension=0, platform_direction=self._platform_direction,
            filter_map=filter_map)

    def _full_column_resolution(
            self,
            row_range: Union[slice, Tuple[int, int, int]],
            col_range: Union[slice, Tuple[int, int, int]],
            filter_map: Optional[numpy.ndarray] = None) -> numpy.ndarray:
        data = super(CSICalculator, self)._full_column_resolution(row_range, col_range)
        return csi_array(
            data, dimension=1, platform_direction=self._platform_direction,
            filter_map=filter_map)

    def _prepare_output(
            self,
            row_range: Union[slice, Tuple[int, int, int]],
            col_range: Union[slice, Tuple[int, int, int]]) -> numpy.ndarray:
        if isinstance(row_range, Sequence):
            row_range = slice(*row_range)
        if isinstance(col_range, Sequence):
            col_range = slice(*col_range)
        row_count = get_slice_result_size(row_range)
        col_count = get_slice_result_size(col_range)
        out_size = (row_count, col_count, 3)
        return numpy.zeros(out_size, dtype=numpy.float64)

    def __getitem__(self, item) -> numpy.ndarray:
        """
        Fetches the csi data based on the input slice.

        Parameters
        ----------
        item

        Returns
        -------
        numpy.ndarray
        """

        if self._fill is None:
            raise ValueError('Unable to proceed unless the index and dimension are set.')

        def get_dimension_details(the_range: Union[slice, Tuple[int, int, int]]):
            if isinstance(the_range, Sequence):
                start, stop, step = the_range
            elif isinstance(the_range, slice):
                start = the_range.start
                stop = the_range.stop
                step = the_range.step
            else:
                raise TypeError('Got unexpected range input {}'.format(the_range))

            full_count = abs(int(stop - start))
            the_snip = -1 if step < 0 else 1
            t_filter_map = filter_map_construction(full_count/self.fill)
            t_block_size = self.get_fetch_block_size(start, stop)
            t_full_range = (start, stop, the_snip)
            return t_filter_map, t_block_size, t_full_range

        # parse the slicing to ensure consistent structure
        row_range, col_range, _ = self._parse_slicing(item)

        if self.dimension == 0:
            # we will proceed fetching full row resolution
            filter_map, row_block_size, this_row_range = get_dimension_details(row_range)
            # get our block definitions
            column_blocks, result_blocks = self.extract_blocks(col_range, row_block_size)
            # noinspection PyTypeChecker
            if len(column_blocks) == 1:
                # it's just a single block
                csi = self._full_row_resolution(this_row_range, col_range, filter_map)
                return csi[::abs(row_range.step), :, :]
            else:
                # prepare the output space
                out = self._prepare_output(row_range, col_range)
                # noinspection PyTypeChecker
                for this_column_range, result_range in zip(column_blocks, result_blocks):
                    csi = self._full_row_resolution(this_row_range, this_column_range, filter_map)
                    out[:, result_range[0]:result_range[1], :] = csi[::abs(row_range.step), :, :]
                return out
        else:
            # we will proceed fetching full column resolution
            filter_map, column_block_size, this_col_range = get_dimension_details(col_range)
            # get our block definitions
            row_blocks, result_blocks = self.extract_blocks(row_range, column_block_size)
            # noinspection PyTypeChecker
            if len(row_blocks) == 1:
                # it's just a single block
                csi = self._full_column_resolution(row_range, this_col_range, filter_map)
                return csi[:, ::abs(col_range.step), :]
            else:
                # prepare the output space
                out = self._prepare_output(row_range, col_range)
                # noinspection PyTypeChecker
                for this_row_range, result_range in zip(row_blocks, result_blocks):
                    csi = self._full_column_resolution(this_row_range, this_col_range, filter_map)
                    out[result_range[0]:result_range[1], :, :] = csi[:, ::abs(col_range.step), :]
                return out



"""
**This sub-package is a work in progress to encapsulate pythonic object-oriented SICD structure 1.1 (2014-09-30).**

This purpose of doing it this way is to encourage effective documentation and streamlined use of the SICD information.
This provides more robustness than using structures with no built-in validation, and more flexibility than using the
rigidity of C++ based standards validation.
"""

__classification__ = "UNCLASSIFIED"



"""
The PositionType definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


from typing import List, Union, Optional

import numpy

from sarpy.io.xml.base import Serializable, SerializableArray
from sarpy.io.xml.descriptors import SerializableDescriptor, SerializableArrayDescriptor

from .base import DEFAULT_STRICT
from .blocks import XYZType, XYZPolyType, XYZPolyAttributeType


class PositionType(Serializable):
    """The details for platform and ground reference positions as a function of time since collection start."""
    _fields = ('ARPPoly', 'GRPPoly', 'TxAPCPoly', 'RcvAPC')
    _required = ('ARPPoly',)
    _collections_tags = {'RcvAPC': {'array': True, 'child_tag': 'RcvAPCPoly'}}

    # descriptors
    ARPPoly = SerializableDescriptor(
        'ARPPoly', XYZPolyType, _required, strict=DEFAULT_STRICT,
        docstring='Aperture Reference Point (ARP) position polynomial in ECF as a function of elapsed '
                  'seconds since start of collection.')  # type: XYZPolyType
    GRPPoly = SerializableDescriptor(
        'GRPPoly', XYZPolyType, _required, strict=DEFAULT_STRICT,
        docstring='Ground Reference Point (GRP) position polynomial in ECF as a function of elapsed '
                  'seconds since start of collection.')  # type: XYZPolyType
    TxAPCPoly = SerializableDescriptor(
        'TxAPCPoly', XYZPolyType, _required, strict=DEFAULT_STRICT,
        docstring='Transmit Aperture Phase Center (APC) position polynomial in ECF as a function of '
                  'elapsed seconds since start of collection.')  # type: XYZPolyType
    RcvAPC = SerializableArrayDescriptor(
        'RcvAPC', XYZPolyAttributeType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Receive Aperture Phase Center polynomials array. '
                  'Each polynomial has output in ECF, and represents a function of elapsed seconds since start of '
                  'collection.')  # type: Union[SerializableArray, List[XYZPolyAttributeType]]

    def __init__(
            self,
            ARPPoly: XYZPolyType = None,
            GRPPoly: Optional[XYZPolyType] = None,
            TxAPCPoly: Optional[XYZPolyType] = None,
            RcvAPC=None,
            **kwargs):
        """

        Parameters
        ----------
        ARPPoly : XYZPolyType
        GRPPoly : None|XYZPolyType
        TxAPCPoly : None|XYZPolyType
        RcvAPC : SerializableArray|List[XYZPolyAttributeType]|list|tuple
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.ARPPoly = ARPPoly
        self.GRPPoly = GRPPoly
        self.TxAPCPoly = TxAPCPoly
        self.RcvAPC = RcvAPC
        super(PositionType, self).__init__(**kwargs)

    def _derive_arp_poly(self, SCPCOA):
        """
        Expected to be called from SICD parent. Set the aperture position polynomial from position, time,
        acceleration at scptime, if necessary.

        .. Note::

            This assumes constant velocity and acceleration.

        Parameters
        ----------
        SCPCOA : sarpy.io.complex.sicd_elements.SCPCOA.SCPCOAType

        Returns
        -------
        None
        """

        if self.ARPPoly is not None:
            return  # nothing to be done

        if SCPCOA is None or SCPCOA.ARPPos is None or SCPCOA.ARPVel is None or SCPCOA.SCPTime is None:
            return  # not enough information to derive

        if SCPCOA.ARPAcc is None:
            SCPCOA.ARPAcc = XYZType.from_array((0, 0, 0))
        # define the polynomial
        coefs = numpy.zeros((3, 3), dtype=numpy.float64)
        scptime = SCPCOA.SCPTime
        pos = SCPCOA.ARPPos.get_array()
        vel = SCPCOA.ARPVel.get_array()
        acc = SCPCOA.ARPAcc.get_array()
        coefs[:, 0] = pos - vel*scptime + 0.5*acc*scptime*scptime
        coefs[:, 1] = vel - acc*scptime
        coefs[:, 2] = acc
        self.ARPPoly = XYZPolyType(X=coefs[0, :], Y=coefs[1, :], Z=coefs[2, :])

    def _basic_validity_check(self) -> bool:
        condition = super(PositionType, self)._basic_validity_check()
        if self.ARPPoly is not None and \
                (self.ARPPoly.X.order1 < 1 or self.ARPPoly.Y.order1 < 1 or self.ARPPoly.Z.order1 < 1):
            self.log_validity_error(
                'ARPPoly should be order at least 1 in each component. '
                'Got X.order1 = {}, Y.order1 = {}, and Z.order1 = {}'.format(self.ARPPoly.X.order1,
                                                                             self.ARPPoly.Y.order1,
                                                                             self.ARPPoly.Z.order1))
            condition = False
        return condition



"""
Base annotation types for general use - based on the geojson implementation
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

import logging
import os
from uuid import uuid4
from typing import Optional, Dict, List, Any, Union
import json

from collections import OrderedDict

from sarpy.geometry.geometry_elements import Jsonable, FeatureCollection, Feature, \
    GeometryCollection, GeometryObject, Geometry, basic_assemble_from_collection


_BASE_VERSION = "Base:1.0"

logger = logging.getLogger(__name__)


class GeometryProperties(Jsonable):
    __slots__ = ('_uid', '_name', '_color')
    _type = 'GeometryProperties'

    def __init__(self, uid=None, name=None, color=None):
        self._name = None
        self._color = None

        if uid is None:
            uid = str(uuid4())
        if not isinstance(uid, str):
            raise TypeError('uid must be a string')
        self._uid = uid

        self.name = name
        self.color = color

    @property
    def uid(self):
        """
        str: A unique identifier for the associated geometry element
        """

        return self._uid

    @property
    def name(self):
        """
        Optional[str]: The name
        """

        return self._name

    @name.setter
    def name(self, value):
        if value is None or isinstance(value, str):
            self._name = value
        else:
            raise TypeError('Got unexpected type for name')

    @property
    def color(self):
        """
        Optional[str]: The color
        """

        return self._color

    @color.setter
    def color(self, value):
        if value is None or isinstance(value, str):
            self._color = value
        else:
            raise TypeError('Got unexpected type for color')

    @classmethod
    def from_dict(cls, the_json):
        """
        Deserialize from json.

        Parameters
        ----------
        the_json : Dict

        Returns
        -------
        GeometryProperties
        """

        typ = the_json['type']
        if typ != cls._type:
            raise ValueError('GeometryProperties cannot be constructed from {}'.format(the_json))
        return cls(
            uid=the_json.get('uid', None),
            name=the_json.get('name', None),
            color=the_json.get('color', None))

    def to_dict(self, parent_dict=None):
        """
        Serialize to json.

        Parameters
        ----------
        parent_dict : None|Dict

        Returns
        -------
        Dict
        """

        if parent_dict is None:
            parent_dict = OrderedDict()
        parent_dict['type'] = self.type
        parent_dict['uid'] = self.uid
        if self.name is not None:
            parent_dict['name'] = self.name
        if self.color is not None:
            parent_dict['color'] = self.color
        return parent_dict


class AnnotationProperties(Jsonable):
    """
    The basic common properties for an annotation
    """

    __slots__ = ('_name', '_description', '_directory', '_geometry_properties', '_parameters')
    _type = 'AnnotationProperties'

    def __init__(self, name=None, description=None, directory=None,
                 geometry_properties=None, parameters=None):
        """

        Parameters
        ----------
        name : Optional[str]
        description : Optional[str]
        directory : Optional[str]
        geometry_properties : None|List[GeometryProperties]
        parameters : Optional[Jsonable]
        """

        self._name = None
        self._description = None
        self._directory = None
        self._geometry_properties = []
        self._parameters = None

        self.name = name
        self.description = description
        self.directory = directory
        self.geometry_properties = geometry_properties
        self.parameters = parameters

    @property
    def name(self):
        """
        Optional[str]: The name
        """
        return self._name

    @name.setter
    def name(self, value):
        if value is None or isinstance(value, str):
            self._name = value
        else:
            raise TypeError('Got unexpected type for name')

    @property
    def description(self):
        """
        Optional[str]: The description
        """
        return self._description

    @description.setter
    def description(self, value):
        if value is None or isinstance(value, str):
            self._description = value
        else:
            raise TypeError('Got unexpected type for description')

    @property
    def directory(self):
        """
        Optional[str]: The directory - for basic display and/or subdivision purposes
        """
        return self._directory

    @directory.setter
    def directory(self, value):
        if value is None:
            self._directory = None
            return

        if not isinstance(value, str):
            raise TypeError('Got unexpected type for directory')

        parts = [entry.strip() for entry in value.split('/')]
        self._directory = '/'.join([entry for entry in parts if entry != ''])

    @property
    def geometry_properties(self):
        # type: () -> List[GeometryProperties]
        """
        List[GeometryProperties]: The geometry properties.
        """

        return self._geometry_properties

    @geometry_properties.setter
    def geometry_properties(self, value):
        if value is None:
            self._geometry_properties = []
            return
        if not isinstance(value, list):
            raise TypeError('Got unexpected value for geometry properties')

        self._geometry_properties = []
        for entry in value:
            self.add_geometry_property(entry)

    def add_geometry_property(self, entry):
        """
        Add a geometry property to the list.

        .. warning::

            Care should be taken that this list stay in sync with the parent geometry.

        Parameters
        ----------
        entry: Dict|GeometryProperties
            The geometry properties instance of serialized version of it.
        """

        if isinstance(entry, dict):
            entry = GeometryProperties.from_dict(entry)

        if not isinstance(entry, GeometryProperties):
            raise TypeError('Got entry of unexpected type for geometry properties list')
        self._geometry_properties.append(entry)

    def get_geometry_property(self, item):
        """
        Fetches the appropriate geometry property.

        Parameters
        ----------
        item : int|str
            The geometry properties uid or integer index.

        Returns
        -------
        GeometryProperties

        Raises
        ------
        KeyError
        """

        return self.get_geometry_property_and_index(item)[0]

    def get_geometry_property_and_index(self, item):
        """
        Fetches the appropriate geometry property and its integer index.

        Parameters
        ----------
        item : int|str
            The geometry properties uid or integer index.

        Returns
        -------
        (GeometryProperties, int)

        Raises
        ------
        KeyError
        """
        if isinstance(item, int):
            return self._geometry_properties[item], item
        elif isinstance(item, str):
            for index, entry in enumerate(self.geometry_properties):
                if entry.uid == item:
                    return entry, index
        raise KeyError('Got unrecognized geometry key `{}`'.format(item))

    @property
    def parameters(self):
        """
        Optional[Jsonable]: The parameters
        """

        return self._parameters

    @parameters.setter
    def parameters(self, value):
        if value is None or isinstance(value, Jsonable):
            self._parameters = value
        else:
            raise TypeError('Got unexpected type for parameters')

    @classmethod
    def from_dict(cls, the_json):
        """
        Deserialize from json.

        Parameters
        ----------
        the_json : Dict

        Returns
        -------
        AnnotationProperties
        """

        typ = the_json['type']
        if typ != cls._type:
            raise ValueError('AnnotationProperties cannot be constructed from {}'.format(the_json))
        return cls(
            name=the_json.get('name', None),
            description=the_json.get('description', None),
            directory=the_json.get('directory', None),
            geometry_properties=the_json.get('geometry_properties', None),
            parameters=the_json.get('parameters', None))

    def to_dict(self, parent_dict=None):
        """
        Serialize to json.

        Parameters
        ----------
        parent_dict : None|Dict

        Returns
        -------
        Dict
        """

        if parent_dict is None:
            parent_dict = OrderedDict()
        parent_dict['type'] = self.type
        for field in ['name', 'description', 'directory']:
            value = getattr(self, field)
            if value is not None:
                parent_dict[field] = value

        if self.geometry_properties is not None:
            parent_dict['geometry_properties'] = [entry.to_dict() for entry in self.geometry_properties]
        if self.parameters is not None:
            parent_dict['parameters'] = self.parameters.to_dict()
        return parent_dict

    def replicate(self):
        geom_properties = None if self.geometry_properties is None else \
            [entry.replicate() for entry in self.geometry_properties]
        params = None if self.parameters is None else self.parameters.replicate()

        the_type = self.__class__
        return the_type(
            name=self.name, description=self.description, directory=self.directory,
            geometry_properties=geom_properties, parameters=params)


class AnnotationFeature(Feature):
    """
    An extension of the Feature class which has the properties attribute
    populated with AnnotationProperties instance.
    """
    _allowed_geometries = None

    @property
    def properties(self):
        """
        The properties.

        Returns
        -------
        None|AnnotationProperties
        """

        return self._properties

    @properties.setter
    def properties(self, properties):
        if properties is None:
            self._properties = AnnotationProperties()
        elif isinstance(properties, AnnotationProperties):
            self._properties = properties
        elif isinstance(properties, dict):
            self._properties = AnnotationProperties.from_dict(properties)
        else:
            raise TypeError('Got an unexpected type for properties attribute of class {}'.format(self.__class__))

    def get_name(self):
        """
        Gets a useful name.

        Returns
        -------
        str
        """

        if self.properties is None or self.properties.name is None:
            return self.uid

        return self.properties.name

    @property
    def geometry(self):
        """
        The geometry object.

        Returns
        -------
        GeometryObject|GeometryCollection
        """

        return self._geometry

    @geometry.setter
    def geometry(self, geometry):
        if isinstance(geometry, dict):
            geometry = Geometry.from_dict(geometry)
        if geometry is None:
            self._geometry = None
            return
        if not isinstance(geometry, Geometry):
            raise TypeError('geometry must be an instance of Geometry, got `{}`'.format(type(geometry)))
        if geometry.is_collection:
            geometry = basic_assemble_from_collection(geometry)
        self._geometry = self._validate_geometry_element(geometry)

    @property
    def geometry_count(self):
        """
        int: The number of base geometry elements
        """

        if self.geometry is None:
            return 0
        elif not self.geometry.is_collection:
            return 1
        else:
            return len(self.geometry.collection)

    def get_geometry_name(self, item):
        """
        Gets the name, or a reasonable default, for the geometry.

        Parameters
        ----------
        item : int|str

        Returns
        -------
        str
        """

        geometry, geom_properties = self.get_geometry_and_geometry_properties(item)
        return '<{}>'.format(geometry.__class__.__name__) if geom_properties.name is None else geom_properties.name

    def get_geometry_property(self, item):
        """
        Gets the geometry properties object for the given index/uid.

        Parameters
        ----------
        item : int|str
            The geometry properties uid or integer index.

        Returns
        -------
        GeometryProperties

        Raises
        ------
        KeyError
        """

        return self.properties.get_geometry_property(item)

    def get_geometry_property_and_index(self, item):
        """
        Gets the geometry properties object and integer index for the given index/uid.

        Parameters
        ----------
        item : int|str
            The geometry properties uid or integer index.

        Returns
        -------
        (GeometryProperties, int)

        Raises
        ------
        KeyError
        """

        return self.properties.get_geometry_property_and_index(item)

    def get_geometry_and_geometry_properties(self, item):
        """
        Gets the geometry and geometry properties object for the given index/uid.

        Parameters
        ----------
        item : int|str
            The geometry properties uid or integer index.

        Returns
        -------
        (Point|Line|Polygon, GeometryProperties)

        Raises
        ------
        KeyError
        """

        if self.geometry is None:
            raise ValueError('No geometry defined.')

        geom_prop, index = self.get_geometry_property_and_index(item)

        index = int(index)
        if not (0 <= index < self.geometry_count):
            raise KeyError('invalid geometry index')
        if self.geometry.is_collection:
            return self.geometry.collection[index], geom_prop
        else:
            return self.geometry, geom_prop

    def get_geometry_element(self, item):
        """
        Gets the basic geometry object at the given index.

        Parameters
        ----------
        item : int|str
            The integer index or associated geometry properties uid.

        Returns
        -------
        Point|Line|Polygon

        Raises
        ------
        ValueError|KeyError
        """

        return self.get_geometry_and_geometry_properties(item)[0]

    def _validate_geometry_element(self, geometry):
        if geometry is None:
            return geometry
        if not isinstance(geometry, Geometry):
            raise TypeError('geometry must be an instance of Geometry base class')

        if self._allowed_geometries is not None and geometry.__class__ not in self._allowed_geometries:
            raise TypeError('geometry ({}) is not of one of the allowed types ({})'.format(geometry, self._allowed_geometries))
        return geometry

    def add_geometry_element(self, geometry, properties=None):
        """
        Adds the given geometry to the feature geometry (collection).

        Parameters
        ----------
        geometry : GeometryObject
        properties : None|GeometryProperties
        """

        if not isinstance(geometry, GeometryObject):
            raise TypeError('geometry must be a GeometryObject instance')

        if properties is None:
            properties = GeometryProperties()
        if not isinstance(properties, GeometryProperties):
            raise TypeError('properties must be a GeometryProperties instance')
        if self.properties is None:
            self.properties = AnnotationProperties()

        # handle the geometry
        self._geometry = self._validate_geometry_element(
            basic_assemble_from_collection(self.geometry, geometry))

        # add the geometry property
        self.properties.add_geometry_property(properties)

        # check that they are in sync
        if len(self.properties.geometry_properties) != self.geometry_count:
            logger.warning(
                'There are {} geometry elements defined\n\t'
                'and {} geometry properties populated. '
                'This is likely to cause problems.'.format(
                    self.geometry_count, len(self.properties.geometry_properties)))

    def remove_geometry_element(self, item):
        """
        Remove the geometry element at the given index

        Parameters
        ----------
        item : int|str
        """

        _, index = self.get_geometry_property_and_index(item)

        if self.geometry_count == 1:
            self.geometry = None
            self.properties = None
        elif self.geometry_count == 2:
            del self.geometry.collection[index]
            del self.properties.geometry_properties[index]
            self.geometry = self.geometry.collection[0]
        else:
            del self.geometry.collection[index]
            del self.properties.geometry_properties[index]


class AnnotationCollection(FeatureCollection):
    """
    An extension of the FeatureCollection class which has the features are
    AnnotationFeature instances.
    """

    @property
    def features(self):
        """
        The features list.

        Returns
        -------
        List[AnnotationFeature]
        """

        return self._features

    @features.setter
    def features(self, features):
        if features is None:
            self._features = None
            self._feature_dict = None
            return

        if not isinstance(features, list):
            raise TypeError('features must be a list of AnnotationFeatures. Got {}'.format(type(features)))

        for entry in features:
            self.add_feature(entry)

    def add_feature(self, feature):
        """
        Add an annotation.

        Parameters
        ----------
        feature : AnnotationFeature|Dict
        """

        if isinstance(feature, dict):
            feature = AnnotationFeature.from_dict(feature)
        if not isinstance(feature, AnnotationFeature):
            raise TypeError('This requires an AnnotationFeature instance, got {}'.format(type(feature)))

        if self._features is None:
            self._feature_dict = {feature.uid: 0}
            self._features = [feature, ]
        else:
            self._feature_dict[feature.uid] = len(self._features)
            self._features.append(feature)

    def __getitem__(self, item):
        # type: (Any) -> Union[AnnotationFeature, List[AnnotationFeature]]
        if self._features is None:
            raise StopIteration

        if isinstance(item, str):
            index = self._feature_dict[item]
            return self._features[index]
        return self._features[item]


class FileAnnotationCollection(Jsonable):
    """
    An collection of annotation elements associated with a given single image element file.
    """

    __slots__ = (
         '_version', '_image_file_name', '_image_id', '_core_name', '_annotations')
    _type = 'FileAnnotationCollection'

    def __init__(self, version=None, annotations=None, image_file_name=None, image_id=None, core_name=None):
        if version is None:
            version = _BASE_VERSION
        self._version = version
        self._annotations = None

        if image_file_name is None:
            self._image_file_name = None
        elif isinstance(image_file_name, str):
            self._image_file_name = os.path.split(image_file_name)[1]
        else:
            raise TypeError('image_file_name must be a None or a string')

        self._image_id = image_id
        self._core_name = core_name

        if self._image_file_name is None and self._image_id is None and self._core_name is None:
            logger.error('One of image_file_name, image_id, or core_name should be defined.')

        self.annotations = annotations

    @property
    def version(self):
        """
        str: The version
        """

        return self._version

    @property
    def image_file_name(self):
        """
        The image file name, if appropriate.

        Returns
        -------
        None|str
        """

        return self._image_file_name

    @property
    def image_id(self):
        """
        The image id, if appropriate.

        Returns
        -------
        None|str
        """

        return self._image_id

    @property
    def core_name(self):
        """
        The image core name, if appropriate.

        Returns
        -------
        None|str
        """

        return self._core_name

    @property
    def annotations(self):
        """
        The annotations.

        Returns
        -------
        AnnotationCollection
        """

        return self._annotations

    @annotations.setter
    def annotations(self, annotations):
        # type: (Union[None, AnnotationCollection, Dict]) -> None
        if annotations is None:
            self._annotations = None
            return

        if isinstance(annotations, AnnotationCollection):
            self._annotations = annotations
        elif isinstance(annotations, dict):
            self._annotations = AnnotationCollection.from_dict(annotations)
        else:
            raise TypeError(
                'annotations must be an AnnotationCollection. Got type {}'.format(type(annotations)))

    def add_annotation(self, annotation):
        """
        Add an annotation.

        Parameters
        ----------
        annotation : AnnotationFeature
            The prospective annotation.
        """

        if isinstance(annotation, dict):
            annotation = AnnotationFeature.from_dict(annotation)
        if not isinstance(annotation, AnnotationFeature):
            raise TypeError('This requires an AnnotationFeature instance. Got {}'.format(type(annotation)))

        if self._annotations is None:
            self._annotations = AnnotationCollection()

        self._annotations.add_feature(annotation)

    def delete_annotation(self, annotation_id):
        """
        Deletes the annotation associated with the given id.

        Parameters
        ----------
        annotation_id : str
        """

        del self._annotations[annotation_id]

    @classmethod
    def from_file(cls, file_name):
        """
        Read from (json) file.

        Parameters
        ----------
        file_name : str

        Returns
        -------
        FileAnnotationCollection
        """

        with open(file_name, 'r') as fi:
            the_dict = json.load(fi)
        return cls.from_dict(the_dict)

    @classmethod
    def from_dict(cls, the_dict):
        """
        Define from a dictionary representation.

        Parameters
        ----------
        the_dict : dict

        Returns
        -------
        FileAnnotationCollection
        """

        if not isinstance(the_dict, dict):
            raise TypeError('This requires a dict. Got type {}'.format(type(the_dict)))

        typ = the_dict.get('type', 'NONE')
        if typ != cls._type:
            raise ValueError('FileAnnotationCollection cannot be constructed from the input dictionary')

        return cls(
            version=the_dict.get('version', 'UNKNOWN'),
            annotations=the_dict.get('annotations', None),
            image_file_name=the_dict.get('image_file_name', None),
            image_id=the_dict.get('image_id', None),
            core_name=the_dict.get('core_name', None))

    def to_dict(self, parent_dict=None):
        if parent_dict is None:
            parent_dict = OrderedDict()
        parent_dict['type'] = self.type
        parent_dict['version'] = self.version
        if self.image_file_name is not None:
            parent_dict['image_file_name'] = self.image_file_name
        if self.image_id is not None:
            parent_dict['image_id'] = self.image_id
        if self.core_name is not None:
            parent_dict['core_name'] = self.core_name
        if self.annotations is not None:
            parent_dict['annotations'] = self.annotations.to_dict()
        return parent_dict

    def to_file(self, file_name):
        with open(file_name, 'w') as fi:
            json.dump(self.to_dict(), fi, indent=1)




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class OBJ(TREElement):
    def __init__(self, value):
        super(OBJ, self).__init__()
        self.add_field('OBJ_TY', 's', 20, value)
        self.add_field('OBJ_NM', 's', 15, value)
        self.add_field('OBJ_POS', 's', 2, value)
        self.add_field('OBJ_SN', 's', 10, value)
        self.add_field('OBJ_LL', 's', 21, value)
        self.add_field('OBJ_ELEV', 's', 8, value)
        self.add_field('OBJ_ROW', 's', 8, value)
        self.add_field('OBJ_COL', 's', 8, value)
        self.add_field('OBJ_PROW', 's', 8, value)
        self.add_field('OBJ_PCOL', 's', 8, value)
        self.add_field('OBJ_ATTR', 's', 20, value)
        self.add_field('OBJ_SEN', 's', 2, value)
        if self.OBJ_SEN == 'R':
            self.add_field('OBJ_AZ_3DB_WIDTH', 's', 7, value)
            self.add_field('OBJ_RNG_3DB_WIDTH', 's', 7, value)
            self.add_field('OBJ_AZ_18DB_WIDTH', 's', 7, value)
            self.add_field('OBJ_RNG_18DB_WIDTH', 's', 7, value)
            self.add_field('OBJ_AZ_3_18DB_RATIO', 's', 8, value)
            self.add_field('OBJ_RNG_3_18DB_RATIO', 's', 8, value)
            self.add_field('OBJ_AZ_PK_SL_RATIO', 's', 8, value)
            self.add_field('OBJ_RNG_PK_SL_RATIO', 's', 8, value)
            self.add_field('OBJ_AZ_INT_SL_RATIO', 's', 8, value)
            self.add_field('OBJ_RNGINT_SL_RATIO', 's', 8, value)
        elif self.OBJ_SEN in ['EO', 'IR']:
            self.add_field('OBJ_CAL_TEMP', 's', 6, value)


class OBJCTAType(TREElement):
    def __init__(self, value):
        super(OBJCTAType, self).__init__()
        self.add_field('VERNUM', 's', 4, value)
        self.add_field('NUM_OBJ', 'd', 3, value)
        self.add_field('OBJ_REF', 's', 10, value)
        self.add_field('NUM_SCENE_OBJ', 'd', 3, value)
        self.add_loop('OBJs', self.NUM_OBJ, OBJ, value)


class OBJCTA(TREExtension):
    _tag_value = 'OBJCTA'
    _data_type = OBJCTAType



"""
A module for performing a selection of validation checks on a SICD (nitf) file,
or the xml file containing the sicd structure.

Use the `check_file` function directly, or perform using the command line

>>> python -m sarpy.consistency.sicd_consistency <file_name>

For more information, about command line usage, see

>>> python -m sarpy.consistency.sicd_consistency --help

"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

import logging
import sys
import argparse
import os
from typing import Union

from sarpy.io.xml.base import parse_xml_from_string, validate_xml_from_string
from sarpy.io.general.nitf import NITFDetails
from sarpy.io.general.nitf_elements.des import DataExtensionHeader, \
    DataExtensionHeader0
from sarpy.io.complex.sicd import SICDReader, SICDDetails, extract_clas
from sarpy.io.complex.sicd_elements.SICD import SICDType
from sarpy.io.complex.sicd_schema import get_urn_details, get_schema_path, \
    get_specification_identifier


logger = logging.getLogger('validation')


def evaluate_xml_versus_schema(xml_string, urn_string):
    """
    Check validity of the xml string versus the appropriate schema.

    Parameters
    ----------
    xml_string : str|bytes
    urn_string : str

    Returns
    -------
    None|bool
    """
    try:
        the_schema = get_schema_path(urn_string)
    except KeyError:
        logger.exception('SICD: Failed getting the schema for urn {}'.format(urn_string))
        return False

    try:
        return validate_xml_from_string(xml_string, the_schema, output_logger=logger)
    except ImportError:
        return None


def _evaluate_xml_string_validity(xml_string):
    """
    Check the validity of the SICD xml, as defined by the given string.

    Parameters
    ----------
    xml_string : str|bytes

    Returns
    -------
    (bool, str, SICDType)
    """

    root_node, xml_ns = parse_xml_from_string(xml_string)
    if xml_ns is None:
        raise ValueError(
            'SICD XML invalid, because no apparent namespace defined in the xml,\n\t'
            'which starts `{}...`'.format(xml_string[:15]))

    if 'default' not in xml_ns:
        raise ValueError(
            'Could not properly interpret the namespace collection from xml\n{}'.format(xml_ns))

    sicd_urn = xml_ns['default']
    # check that our urn is mapped
    try:
        _ = get_urn_details(sicd_urn)
        check_schema = True
    except Exception as e:
        logger.exception('SICD: The SICD namespace has unrecognized value')
        check_schema = False

    valid_xml = None
    if check_schema:
        valid_xml = evaluate_xml_versus_schema(xml_string, sicd_urn)
    if valid_xml is None:
        valid_xml = True

    # perform the various sicd structure checks
    the_sicd = SICDType.from_node(root_node, xml_ns=xml_ns)
    valid_sicd_contents = the_sicd.is_valid(recursive=True, stack=False)
    return valid_xml & valid_sicd_contents, sicd_urn, the_sicd


def check_sicd_data_extension(nitf_details, des_header, xml_string):
    """
    Evaluate a SICD data extension for validity.

    Parameters
    ----------
    nitf_details : NITFDetails
    des_header : DataExtensionHeader|DataExtensionHeader0
    xml_string : str|bytes

    Returns
    -------
    (bool, SICDType)
    """

    def check_des_header_fields():
        # type: () -> bool

        des_id = des_header.DESID.strip() if nitf_details.nitf_version == '02.10' else des_header.DESTAG.strip()

        if des_id != 'XML_DATA_CONTENT':
            logger.warning('SICD: Found old style SICD DES Header. This is deprecated.')
            return True

        # make sure that the NITF urn is evaluated for sensibility
        nitf_urn = des_header.UserHeader.DESSHTN.strip()
        try:
            nitf_urn_details = get_urn_details(nitf_urn)
        except Exception:
            logger.exception('SICD: The SICD DES.DESSHTN must be a recognized urn')
            return False

        # make sure that the NITF urn and SICD urn actually agree
        header_good = True
        if nitf_urn != xml_urn:
            logger.error('SICD: The SICD DES.DESSHTN ({}) and urn ({}) must agree'.format(nitf_urn, xml_urn))
            header_good = False

        # make sure that the NITF DES fields are populated appropriately for NITF urn
        if des_header.UserHeader.DESSHSI.strip() != get_specification_identifier():
            logger.error(
                'SICD: DES.DESSHSI has value `{}`,\n\tbut should have value `{}`'.format(
                    des_header.UserHeader.DESSHSI.strip(), get_specification_identifier()))
            header_good = False

        nitf_version = nitf_urn_details['version']
        if des_header.UserHeader.DESSHSV.strip() != nitf_version:
            logger.error(
                'SICD: DES.DESSHSV has value `{}`,\n\tbut should have value `{}` based on DES.DESSHTN `{}`'.format(
                    des_header.UserHeader.DESSHSV.strip(), nitf_version, nitf_urn))
            header_good = False

        nitf_date = nitf_urn_details['date']
        if des_header.UserHeader.DESSHSD.strip() != nitf_date:
            logger.warning(
                'SICD: DES.DESSHSD has value `{}`,\n\tbut should have value `{}` based on DES.DESSHTN `{}`'.format(
                    des_header.UserHeader.DESSHSD.strip(), nitf_date, nitf_urn))
        return header_good

    def compare_sicd_class():
        # type: () -> bool

        if the_sicd.CollectionInfo is None or the_sicd.CollectionInfo.Classification is None:
            logger.error(
                'SICD: SICD.CollectionInfo.Classification is not populated,\n\t'
                'so can not be compared with SICD DES.DESCLAS `{}`'.format(des_header.Security.CLAS.strip()))
            return False

        sicd_class = the_sicd.CollectionInfo.Classification
        extracted_class = extract_clas(the_sicd)
        if extracted_class != des_header.Security.CLAS.strip():
            logger.warning(
                'SICD: DES.DESCLAS is `{}`,\n\tand SICD.CollectionInfo.Classification '
                'is {}'.format(des_header.Security.CLAS.strip(), sicd_class))

        if des_header.Security.CLAS.strip() != nitf_details.nitf_header.Security.CLAS.strip():
            logger.warning(
                'SICD: DES.DESCLAS is `{}`,\n\tand NITF.CLAS is `{}`'.format(
                    des_header.Security.CLAS.strip(), nitf_details.nitf_header.Security.CLAS.strip()))
        return True

    # check sicd xml structure for validity
    valid_sicd, xml_urn, the_sicd = _evaluate_xml_string_validity(xml_string)
    # check that the sicd information and header information appropriately match
    valid_header = check_des_header_fields()
    # check that the classification seems to make sense
    valid_class = compare_sicd_class()
    return valid_sicd & valid_header & valid_class, the_sicd


def check_sicd_file(nitf_details):
    """
    Check the validity of the given NITF file as a SICD file.

    Parameters
    ----------
    nitf_details : str|NITFDetails
        The path to the NITF file, or a `NITFDetails` object.

    Returns
    -------
    bool
    """

    def check_data_extension_headers():
        # type: () -> (str, Union[DataExtensionHeader, DataExtensionHeader0])
        sicd_des = []

        for i in range(nitf_details.des_subheader_offsets.size):
            subhead_bytes = nitf_details.get_des_subheader_bytes(i)
            des_bytes = None
            if subhead_bytes.startswith(b'DEXML_DATA_CONTENT'):
                des_bytes = nitf_details.get_des_bytes(i)

            elif subhead_bytes.startswith(b'DESIDD_XML'):
                raise ValueError(
                    'This file contains an old format SIDD DES, and should be a SIDD file')
            elif subhead_bytes.startswith(b'DESICD_XML'):
                des_bytes = nitf_details.get_des_bytes(i)

            if des_bytes is None:
                continue

            # compare the SICD structure and the des header structure
            if nitf_details.nitf_version == '02.00':
                des_header = DataExtensionHeader0.from_bytes(subhead_bytes, start=0)
            elif nitf_details.nitf_version == '02.10':
                des_header = DataExtensionHeader.from_bytes(subhead_bytes, start=0)
            else:
                raise ValueError('Got unhandled NITF version {}'.format(nitf_details.nitf_version))

            try:
                des_bytes = des_bytes.decode('utf-8').strip().encode()
                root_node, xml_ns = parse_xml_from_string(des_bytes)
                # namespace makes this ugly
                if 'SIDD' in root_node.tag:
                    raise ValueError(
                        'This file contains a SIDD DES, and should be a SIDD file')
                elif 'SICD' in root_node.tag:
                    sicd_des.append((i, des_bytes, des_header))
            except Exception as e:
                logger.error('Failed parsing the xml DES entry {} as xml'.format(i))
                raise e

        if len(sicd_des) == 0:
            raise ValueError('No SICD DES values found, so this is not a viable SICD file')
        elif len(sicd_des) > 1:
            raise ValueError(
                'Multiple SICD DES values found at indices {},\n'
                'so this is not a viable SICD file'.format([entry[0] for entry in sicd_des]))

        return sicd_des[0][1], sicd_des[0][2]

    def check_image_data():
        # type: () -> bool

        # get pixel type
        pixel_type = the_sicd.ImageData.PixelType
        if pixel_type == 'RE32F_IM32F':
            exp_nbpp = 32
            exp_pvtype = 'R'
        elif pixel_type == 'RE16I_IM16I':
            exp_nbpp = 16
            exp_pvtype = 'SI'
        elif pixel_type == 'AMP8I_PHS8I':
            exp_nbpp = 8
            exp_pvtype = 'INT'
        else:
            raise ValueError('Got unexpected pixel type {}'.format(pixel_type))

        valid_images = True
        # verify that all images have the correct pixel type
        for i, img_header in enumerate(nitf_details.img_headers):
            if img_header.ICAT.strip() != 'SAR':
                valid_images = False
                logger.error(
                    'SICD: image segment at index {} of {} has ICAT = `{}`,\n\texpected to be `SAR`'.format(
                        i, len(nitf_details.img_headers), img_header.ICAT.strip()))

            if img_header.PVTYPE.strip() != exp_pvtype:
                valid_images = False
                logger.error(
                    'SICD: image segment at index {} of {} has PVTYPE = `{}`,\n\t'
                    'expected to be `{}` based on pixel type {}'.format(
                        i, len(nitf_details.img_headers), img_header.PVTYPE.strip(), exp_pvtype, pixel_type))
            if img_header.NBPP != exp_nbpp:
                valid_images = False
                logger.error(
                    'SICD: image segment at index {} of {} has NBPP = `{}`,\n\t'
                    'expected to be `{}` based on pixel type {}'.format(
                        i, len(nitf_details.img_headers), img_header.NBPP, exp_nbpp, pixel_type))

            if len(img_header.Bands) != 2:
                valid_images = False
                logger.error('SICD: image segment at index {} of {} does not have two (I/Q or M/P) bands'.format(
                    i, len(nitf_details.img_headers)))
                continue

            if pixel_type == 'AMP8I_PHS8I':
                if img_header.Bands[0].ISUBCAT.strip() != 'M' and img_header.Bands[1].ISUBCAT.strip() != 'P':
                    valid_images = False
                    logger.error(
                        'SICD: pixel_type is {}, image segment at index {} of {}\n\t'
                        'has bands with ISUBCAT {}, expected ("M", "P")'.format(
                            pixel_type, i, len(nitf_details.img_headers),
                            (img_header.Bands[0].ISUBCAT.strip(), img_header.Bands[1].ISUBCAT.strip())))
            else:
                if img_header.Bands[0].ISUBCAT.strip() != 'I' and img_header.Bands[1].ISUBCAT.strip() != 'Q':
                    valid_images = False
                    logger.error(
                        'SICD: pixel_type is {}, image segment at index {} of {}\n\t'
                        'has bands with ISUBCAT {}, expected ("I", "Q")'.format(
                            pixel_type, i, len(nitf_details.img_headers),
                            (img_header.Bands[0].ISUBCAT.strip(), img_header.Bands[1].ISUBCAT.strip())))

        return valid_images

    if isinstance(nitf_details, str):
        if not os.path.isfile(nitf_details):
            raise ValueError('Got string input, but it is not a valid path')
        nitf_details = NITFDetails(nitf_details)

    if not isinstance(nitf_details, NITFDetails):
        raise TypeError(
            'Input is expected to be a path to a NITF file, or a NITFDetails object instance')

    # find the sicd header
    sicd_xml_string, des_header = check_data_extension_headers()
    # check that the sicd and header are valid
    valid_sicd_des, the_sicd = check_sicd_data_extension(nitf_details, des_header, sicd_xml_string)
    # check that the image segments all make sense compared to the sicd structure
    valid_img = check_image_data()
    all_valid = valid_sicd_des & valid_img

    if valid_img:
        try:
            reader = SICDReader(nitf_details.file_name)
        except Exception as e:
            logger.exception(
                'SICD: All image segments appear viable for the SICD,\n\t'
                'but SICDReader construction failed')
    return all_valid


def check_file(file_name):
    """
    Check the SICD validity for the given file SICD (i.e. appropriately styled NITF)
    or xml file containing the SICD structure alone.

    Parameters
    ----------
    file_name : str|SICDDetails

    Returns
    -------
    bool
    """

    if isinstance(file_name, str):
        if not os.path.isfile(file_name):
            raise ValueError('Got string input, but it is not a valid path')

        # check if this is just an xml file
        with open(file_name, 'rb') as fi:
            initial_bits = fi.read(30)
            if initial_bits.startswith(b'<?xml') or initial_bits.startswith(b'<SICD'):
                fi.seek(0)
                sicd_xml = fi.read()
                return _evaluate_xml_string_validity(sicd_xml)[0]

    return check_sicd_file(file_name)


def main(args=None):
    parser = argparse.ArgumentParser('SICD Consistency')
    parser.add_argument('file_name')
    parser.add_argument(
        '-l', '--level', default='WARNING',
        choices=['INFO', 'WARNING', 'ERROR'], help="Logging level")
    config = parser.parse_args(args)

    logging.basicConfig(level=config.level)
    logger.setLevel(config.level)
    validity = check_file(config.file_name)
    if validity:
        logger.info('\nSICD: {} has been validated with no errors'.format(config.file_name))
    else:
        logger.error('\nSICD: {} has apparent errors'.format(config.file_name))
    return int(validity)


if __name__ == '__main__':
    sys.exit(main())    # pragma: no cover



"""
The PFAType definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from typing import Union, Optional, Tuple

import numpy
from numpy.linalg import norm
from numpy.polynomial import polynomial

from sarpy.io.xml.base import Serializable
from sarpy.io.xml.descriptors import BooleanDescriptor, FloatDescriptor, \
    SerializableDescriptor, UnitVectorDescriptor

from .base import DEFAULT_STRICT, FLOAT_FORMAT
from .blocks import Poly1DType, Poly2DType, XYZType

from sarpy.geometry import geocoords


class STDeskewType(Serializable):
    """
    Parameters to describe image domain ST Deskew processing.
    """

    _fields = ('Applied', 'STDSPhasePoly')
    _required = _fields
    # descriptors
    Applied = BooleanDescriptor(
        'Applied', _required, strict=DEFAULT_STRICT,
        docstring='Parameter indicating if slow time *(ST)* Deskew Phase function has been applied.')  # type: bool
    STDSPhasePoly = SerializableDescriptor(
        'STDSPhasePoly', Poly2DType, _required, strict=DEFAULT_STRICT,
        docstring='Slow time deskew phase function to perform the *ST/Kaz* shift. Two-dimensional phase '
                  '(cycles) polynomial function of image range coordinate *(variable 1)* and '
                  'azimuth coordinate *(variable 2)*.')  # type: Poly2DType

    def __init__(
            self,
            Applied: bool = None,
            STDSPhasePoly: Union[Poly2DType, numpy.ndarray, list, tuple] = None,
            **kwargs):
        """

        Parameters
        ----------
        Applied : bool
        STDSPhasePoly : Poly2DType|numpy.ndarray|list|tuple
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Applied = Applied
        # noinspection PytypeChecker
        self.STDSPhasePoly = STDSPhasePoly
        super(STDeskewType, self).__init__(**kwargs)


class PFAType(Serializable):
    """Parameters for the Polar Formation Algorithm."""
    _fields = (
        'FPN', 'IPN', 'PolarAngRefTime', 'PolarAngPoly', 'SpatialFreqSFPoly', 'Krg1', 'Krg2', 'Kaz1', 'Kaz2',
        'STDeskew')
    _required = ('FPN', 'IPN', 'PolarAngRefTime', 'PolarAngPoly', 'SpatialFreqSFPoly', 'Krg1', 'Krg2', 'Kaz1', 'Kaz2')
    _numeric_format = {
        'PolarAngRefTime': FLOAT_FORMAT, 'Krg1': FLOAT_FORMAT, 'Krg2': FLOAT_FORMAT,
        'Kaz1': FLOAT_FORMAT, 'Kaz2': FLOAT_FORMAT}
    # descriptors
    FPN = UnitVectorDescriptor(
        'FPN', XYZType, _required, strict=DEFAULT_STRICT,
        docstring='Focus Plane unit normal in ECF coordinates. Unit vector FPN points away from the center of '
                  'the Earth.')  # type: XYZType
    IPN = UnitVectorDescriptor(
        'IPN', XYZType, _required, strict=DEFAULT_STRICT,
        docstring='Image Formation Plane unit normal in ECF coordinates. Unit vector IPN points away from the '
                  'center of the Earth.')  # type: XYZType
    PolarAngRefTime = FloatDescriptor(
        'PolarAngRefTime', _required, strict=DEFAULT_STRICT,
        docstring='Polar image formation reference time *(in seconds)*. Polar Angle = 0 at the reference time. '
                  'Measured relative to collection start. *Note: Reference time is typically set equal to the SCP '
                  'COA time but may be different.*')  # type: float
    PolarAngPoly = SerializableDescriptor(
        'PolarAngPoly', Poly1DType, _required, strict=DEFAULT_STRICT,
        docstring='Polynomial function that yields Polar Angle *(in radians)* as function of time '
                  'relative to Collection Start.')  # type: Poly1DType
    SpatialFreqSFPoly = SerializableDescriptor(
        'SpatialFreqSFPoly', Poly1DType, _required, strict=DEFAULT_STRICT,
        docstring='Polynomial that yields the *Spatial Frequency Scale Factor (KSF)* as a function of Polar '
                  r'Angle. That is, :math:`Polar Angle[radians] \to KSF[dimensionless]`. Used to scale RF '
                  'frequency *(fx, Hz)* to aperture spatial frequency *(Kap, cycles/m)*. Where,'
                  r':math:`Kap = fx\cdot (2/c)\cdot KSF`, and `Kap` is the effective spatial '
                  'frequency in the polar aperture.')  # type: Poly1DType
    Krg1 = FloatDescriptor(
        'Krg1', _required, strict=DEFAULT_STRICT,
        docstring='Minimum *range spatial frequency (Krg)* output from the polar to rectangular '
                  'resampling.')  # type: float
    Krg2 = FloatDescriptor(
        'Krg2', _required, strict=DEFAULT_STRICT,
        docstring='Maximum *range spatial frequency (Krg)* output from the polar to rectangular '
                  'resampling.')  # type: float
    Kaz1 = FloatDescriptor(
        'Kaz1', _required, strict=DEFAULT_STRICT,
        docstring='Minimum *azimuth spatial frequency (Kaz)* output from the polar to rectangular '
                  'resampling.')  # type: float
    Kaz2 = FloatDescriptor(
        'Kaz2', _required, strict=DEFAULT_STRICT,
        docstring='Maximum *azimuth spatial frequency (Kaz)* output from the polar to rectangular '
                  'resampling.')  # type: float
    STDeskew = SerializableDescriptor(
        'STDeskew', STDeskewType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters to describe image domain slow time *(ST)* Deskew processing.')  # type: STDeskewType

    def __init__(
            self,
            FPN: Union[XYZType, numpy.ndarray, list, tuple] = None,
            IPN: Union[XYZType, numpy.ndarray, list, tuple] = None,
            PolarAngRefTime: float = None,
            PolarAngPoly: Union[Poly1DType, numpy.ndarray, list, tuple] = None,
            SpatialFreqSFPoly: Union[Poly1DType, numpy.ndarray, list, tuple] = None,
            Krg1: float = None,
            Krg2: float = None,
            Kaz1: float = None,
            Kaz2: float = None,
            STDeskew: Optional[STDeskewType] = None,
            **kwargs):
        """

        Parameters
        ----------
        FPN : XYZType|numpy.ndarray|list|tuple
        IPN : XYZType|numpy.ndarray|list|tuple
        PolarAngRefTime : float
        PolarAngPoly : Poly1DType|numpy.ndarray|list|tuple
        SpatialFreqSFPoly : Poly1DType|numpy.ndarray|list|tuple
        Krg1 : float
        Krg2 : float
        Kaz1 : float
        Kaz2 : float
        STDeskew : STDeskewType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.FPN = FPN
        self.IPN = IPN
        self.PolarAngRefTime = PolarAngRefTime
        self.PolarAngPoly = PolarAngPoly
        self.SpatialFreqSFPoly = SpatialFreqSFPoly
        self.Krg1, self.Krg2 = Krg1, Krg2
        self.Kaz1, self.Kaz2 = Kaz1, Kaz2
        self.STDeskew = STDeskew
        super(PFAType, self).__init__(**kwargs)

    def pfa_polar_coords(
            self,
            Position,
            SCP: numpy.ndarray,
            times: Union[float, int, numpy.ndarray]) -> Tuple[
                Union[None, float, numpy.ndarray], Union[None, float, numpy.ndarray]]:
        """
        Calculate the PFA parameters necessary for mapping phase history to polar coordinates.

        Parameters
        ----------
        Position : sarpy.io.complex.sicd_elements.Position.PositionType
        SCP : numpy.ndarray
        times : numpy.ndarray|float|int

        Returns
        -------
        k_a : None|float|numpy.ndarray
            The polar angle
        k_sf : None|float|numpy.ndarray
            The spatial frequency scale factor. The shape of the return arrays
            will match the shape of the `times` array (or scalar).
        """

        def project_to_image_plane(points):
            # type: (numpy.ndarray) -> numpy.ndarray
            # project into the image plane along line normal to the focus plane
            offset = (SCP - points).dot(ipn)/fpn.dot(ipn)
            if offset.ndim == 0:
                return points + offset*fpn
            else:
                return points + numpy.outer(offset, fpn)

        if self.IPN is None or self.FPN is None:
            return None, None

        ipn = self.IPN.get_array(dtype='float64')
        fpn = self.FPN.get_array(dtype='float64')
        if isinstance(times, (float, int)) or times.ndim == 0:
            o_shape = None
            times = numpy.array([times, ], dtype='float64')
        else:
            o_shape = times.shape
            times = numpy.reshape(times, (-1, ))
        positions = Position.ARPPoly(times)
        reference_position = Position.ARPPoly(self.PolarAngRefTime)
        image_plane_positions = project_to_image_plane(positions)
        image_plane_coa = project_to_image_plane(reference_position)

        # establish image plane coordinate system
        ip_x = image_plane_coa - SCP
        ip_x /= numpy.linalg.norm(ip_x)
        ip_y = numpy.cross(ipn, ip_x)

        # compute polar angle of sensor position in image plane
        ip_range = image_plane_positions - SCP
        ip_range /= numpy.linalg.norm(ip_range, axis=1)[:, numpy.newaxis]
        k_a = numpy.arctan2(ip_range.dot(ip_y), ip_range.dot(ip_x))

        # compute the spatial frequency scale factor
        range_vectors = positions - SCP
        range_vectors /= numpy.linalg.norm(range_vectors, axis=1)[:, numpy.newaxis]
        sin_graze = range_vectors.dot(fpn)
        sin_graze_ip = ip_range.dot(fpn)
        k_sf = numpy.sqrt((1 - sin_graze*sin_graze)/(1 - sin_graze_ip*sin_graze_ip))
        if o_shape is None:
            return k_a[0], k_sf[0]
        elif len(o_shape) > 1:
            return numpy.reshape(k_a, o_shape), numpy.reshape(k_sf, o_shape)
        else:
            return k_a, k_sf

    def _derive_parameters(self, Grid, SCPCOA, GeoData, Position, Timeline):
        """
        Expected to be called from SICD parent.

        Parameters
        ----------
        Grid : sarpy.io.complex.sicd_elements.Grid.GridType
        SCPCOA : sarpy.io.complex.sicd_elements.SCPCOA.SCPCOAType
        GeoData : sarpy.io.complex.sicd_elements.GeoData.GeoDataType
        Position : sarpy.io.complex.sicd_elements.Position.PositionType
        Timeline : sarpy.io.complex.sicd_elements.Timeline.TimelineType

        Returns
        -------
        None
        """

        if self.PolarAngRefTime is None and SCPCOA.SCPTime is not None:
            self.PolarAngRefTime = SCPCOA.SCPTime

        if GeoData is None or GeoData.SCP is None or GeoData.SCP.ECF is None:
            return

        scp = GeoData.SCP.ECF.get_array()

        if SCPCOA.ARPPos is not None and SCPCOA.ARPVel is not None:
            scp = GeoData.SCP.ECF.get_array()
            etp = geocoords.wgs_84_norm(scp)

            arp = SCPCOA.ARPPos.get_array()
            los = (scp - arp)
            ulos = los/norm(los)

            look = SCPCOA.look
            arp_vel = SCPCOA.ARPVel.get_array()
            uspz = look*numpy.cross(arp_vel, ulos)
            uspz /= norm(uspz)
            if Grid is not None and Grid.ImagePlane is not None:
                if self.IPN is None:
                    if Grid.ImagePlane == 'SLANT':
                        self.IPN = XYZType.from_array(uspz)
                    elif Grid.ImagePlane == 'GROUND':
                        self.IPN = XYZType.from_array(etp)
            elif self.IPN is None:
                self.IPN = XYZType.from_array(uspz)  # assuming slant -> most common

            if self.FPN is None:
                self.FPN = XYZType.from_array(etp)

        if Position is not None and \
                Timeline is not None and Timeline.CollectDuration is not None and \
                (self.PolarAngPoly is None or self.SpatialFreqSFPoly is None):
            pol_ref_pos = Position.ARPPoly(self.PolarAngRefTime)
            # fit the PFA polynomials
            times = numpy.linspace(0, Timeline.CollectDuration, 15)
            k_a, k_sf = self.pfa_polar_coords(Position, scp, times)

            self.PolarAngPoly = Poly1DType(Coefs=polynomial.polyfit(times, k_a, 5, full=False))
            self.SpatialFreqSFPoly = Poly1DType(Coefs=polynomial.polyfit(k_a, k_sf, 5, full=False))

        if Grid is not None and Grid.Row is not None and \
                Grid.Row.KCtr is not None and Grid.Row.ImpRespBW is not None:
            if self.Krg1 is None:
                self.Krg1 = Grid.Row.KCtr - 0.5*Grid.Row.ImpRespBW
            if self.Krg2 is None:
                self.Krg2 = Grid.Row.KCtr + 0.5*Grid.Row.ImpRespBW
        if Grid is not None and Grid.Col is not None and \
                Grid.Col.KCtr is not None and Grid.Col.ImpRespBW is not None:
            if self.Kaz1 is None:
                self.Kaz1 = Grid.Col.KCtr - 0.5*Grid.Col.ImpRespBW
            if self.Kaz2 is None:
                self.Kaz2 = Grid.Col.KCtr + 0.5*Grid.Col.ImpRespBW

    def _check_polar_ang_ref(self):
        """
        Checks the polar angle origin makes sense.

        Returns
        -------
        bool
        """

        if self.PolarAngPoly is None or self.PolarAngRefTime is None:
            return True

        cond = True
        polar_angle_ref = self.PolarAngPoly(self.PolarAngRefTime)
        if abs(polar_angle_ref) > 1e-4:
            self.log_validity_error(
                'The PolarAngPoly evaluated at PolarAngRefTime yields {}, which should be 0'.format(polar_angle_ref))
            cond = False
        return cond

    def _basic_validity_check(self) -> bool:
        condition = super(PFAType, self)._basic_validity_check()
        condition &= self._check_polar_ang_ref()
        return condition




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class CSCRNAType(TREElement):
    def __init__(self, value):
        super(CSCRNAType, self).__init__()
        self.add_field('PREDICT_CORNERS', 's', 1, value)
        self.add_field('ULCNR_LAT', 's', 9, value)
        self.add_field('ULCNR_LONG', 's', 10, value)
        self.add_field('ULCNR_HT', 's', 8, value)
        self.add_field('URCNR_LAT', 's', 9, value)
        self.add_field('URCNR_LONG', 's', 10, value)
        self.add_field('URCNR_HT', 's', 8, value)
        self.add_field('LRCNR_LAT', 's', 9, value)
        self.add_field('LRCNR_LONG', 's', 10, value)
        self.add_field('LRCNR_HT', 's', 8, value)
        self.add_field('LLCNR_LAT', 's', 9, value)
        self.add_field('LLCNR_LONG', 's', 10, value)
        self.add_field('LLCNR_HT', 's', 8, value)


class CSCRNA(TREExtension):
    _tag_value = 'CSCRNA'
    _data_type = CSCRNAType



"""
The DataType definition for CPHD 0.3.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from typing import List

from sarpy.io.phase_history.cphd1_elements.base import DEFAULT_STRICT
from sarpy.io.xml.base import Serializable
from sarpy.io.xml.descriptors import StringEnumDescriptor, IntegerDescriptor, SerializableListDescriptor


class ArraySizeType(Serializable):
    """
    Parameters that define the array sizes.
    """

    _fields = ('NumVectors', 'NumSamples')
    _required = ('NumVectors', 'NumSamples')
    # descriptors
    NumVectors = IntegerDescriptor(
        'NumVectors', _required, strict=DEFAULT_STRICT, bounds=(1, None),
        docstring='Number of slow time vectors in the PHD array in this channel.')  # type: int
    NumSamples = IntegerDescriptor(
        'NumSamples', _required, strict=DEFAULT_STRICT, bounds=(1, None),
        docstring='Number of samples per vector in the PHD array in this channel.')  # type: int

    def __init__(self, NumVectors=None, NumSamples=None, **kwargs):
        """

        Parameters
        ----------
        NumVectors : int
        NumSamples : int
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.NumVectors = NumVectors
        self.NumSamples = NumSamples
        super(ArraySizeType, self).__init__(**kwargs)


class DataType(Serializable):
    """
    Parameters that describe binary data components contained in the product.
    """

    _fields = ('SampleType', 'NumCPHDChannels', 'NumBytesVBP', 'ArraySize')
    _required = ('SampleType', 'NumBytesVBP', 'ArraySize')
    _collections_tags = {'ArraySize': {'array': False, 'child_tag': 'ArraySize'}}
    # descriptors
    SampleType = StringEnumDescriptor(
        'SampleType', ("RE32F_IM32F", "RE16I_IM16I", "RE08I_IM08I"), _required, strict=True,
        docstring="Indicates the PHD sample format of the PHD array(s). All arrays "
                  "have the sample type. Real and imaginary components stored in adjacent "
                  "bytes, real component stored first.")  # type: str
    NumBytesVBP = IntegerDescriptor(
        'NumBytesVBP', _required, strict=DEFAULT_STRICT, bounds=(1, None),
        docstring='Number of bytes per set of Vector Based Parameters.')  # type: int
    ArraySize = SerializableListDescriptor(
        'ArraySize', ArraySizeType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='CPHD array size parameters.')  # type: List[ArraySizeType]

    def __init__(self, SampleType=None, NumBytesVBP=None, ArraySize=None, **kwargs):
        """

        Parameters
        ----------
        SampleType : str
        NumBytesVBP : int
        ArraySize : List[ArraySizeType]
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.SampleType = SampleType
        self.NumBytesVBP = NumBytesVBP
        self.ArraySize = ArraySize
        super(DataType, self).__init__(**kwargs)

    @property
    def NumCPHDChannels(self):
        """
        int: The number of CPHD channels.
        """

        if self.ArraySize is None:
            return 0
        return len(self.ArraySize)
